# -*- coding: utf-8 -*-
"""
`FFProbe Wrapper`
~~~~~~~~~~~~~~~~~
>>> import transcend as ts
>>> probe = ts.Probe('https://s3.amazonaws.com/bucket/vid.mp4')
>>> probe.exe()
..
‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
| (Executing)                                                                  |
| ffprobe                                                                      |
|··············································································|
|(Output) ffprobe version git-2016-06-13-1f87233 Copyright (c) 2007-2016 the FF|
|  built with gcc 4.8 (Ubuntu 4.8.4-2ubuntu1~14.04.3)                          |
|  configuration: --enable-gpl --enable-gnutls --enable-openssl --enable-nonfre|
|  libavutil      55. 24.100 / 55. 24.100                                      |
|  libavcodec     57. 46.100 / 57. 46.100                                      |
|  libavformat    57. 38.100 / 57. 38.100                                      |
|  libavdevice    57.  0.101 / 57.  0.101                                      |
|  libavfilter     6. 46.101 /  6. 46.101                                      |
|  libswscale      4.  1.100 /  4.  1.100                                      |
|  libswresample   2.  1.100 /  2.  1.100                                      |
|  libpostproc    54.  0.100 / 54.  0.100                                      |
|Input #0, mov,mp4,m4a,3gp,3g2,mj2, from 'https://s3.amazonaws.com/bucket/vid.m|
|  Metadata:                                                                   |
|    major_brand     : isom                                                    |
|    minor_version   : 512                                                     |
|    compatible_brands: isomiso2avc1mp41                                       |
|    encoder         : Lavf57.25.100                                           |
|  Duration: 00:00:01.60, start: 0.000000, bitrate: 478 kb/s                   |
|    Stream #0:0(und): Audio: aac (LC) (mp4a / 0x6134706D), 32000 Hz, stereo, f|
|    Metadata:                                                                 |
|      handler_name    : SoundHandler                                          |
|    Stream #0:1(und): Video: h264 (Constrained Baseline) (avc1 / 0x31637661), |
|    Metadata:                                                                 |
|      handler_name    : VideoHandler                                          |
|                                                                              |
‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
..

"""
import sys
import subprocess

from .cli import *
from .ffmpeg import Ffmpeg, FfmpegResult


# TODO: https://trac.ffmpeg.org/wiki/FFprobeTips


class Probe(Ffmpeg):
    """
    ``FFProbe wrapper``
    ~~~~~~~~~~~~~~~~~~~
    @format: (#str) Force format to use
    @unit: (#bool) Show the unit of the displayed values.
    @prefix: (#bool) Use SI prefixes for the displayed values.
    @byte_binary_prefix: (#bool) Force the use of binary prefixes for byte
        values.
    @sexagesimal: (#bool) Use sexagesimal format HH:MM:SS.MICROSECONDS for time
        values.
    @pretty: (#bool) Prettify the format of the displayed values, it
        corresponds to the options "-unit -prefix -byte_binary_prefix
        -sexagesimal".
    @print_format: (#str) Set the output printing format
    @sections: (#bool) Print sections structure and section information,
        and exit. The output is not meant to be parsed by a machine.
    @select_streams: (#str) Select only the streams specified by
        |stream_specifier|. This option affects only the options related to
        streams (e.g. show_streams, show_packets, etc.)
    @show_data_hash: (#str) Show a hash of payload data, for packets with
        |-show_packets| and for codec extradata with |-show_streams|
    @show_error: (#bool) Show information about the error found when trying to
        probe the input.
    @show_format: (#bool) Show information about the container format of the
        input multimedia stream.
    @show_packets: (#bool) Show information about each packet contained in the
        input multimedia stream.
    @show_frames: (#bool) Show information about each frame and subtitle
        contained in the input multimedia stream.
    @show_streams: (#bool) Show information about each media stream contained
        in the input multimedia stream.
    @show_chapters: (#bool) Show information about chapters stored in the
        format.
    @count_frames: (#bool) Count the number of frames per stream and report it
        in the corresponding stream section.
    @count_packets: (#bool) Count the number of packets per stream and report
        it in the corresponding stream section.
    @read_interval: (#str) Read only the specified intervals. |read_interval|
        must be a sequence of interval specifications separated by ",".
        ffprobe will seek to the interval starting point, and will continue
        reading from that.
    @private: (#bool) Prevents showing private data
    @show_program_version: (#bool) Show information related to program version.
    @show_library_versions: (#bool) Show information related to library
        versions.
    @show_versions: (#bool) Show information related to program and library
        versions. This is the equivalent of setting both -show_program_version
        and -show_library_versions options.
    @show_pixel_formats: (#bool) Show information about all pixel formats
        supported by FFmpeg.
    @bitexact: (#bool) Force bitexact output, useful to produce output which
        is not dependent on the specific build.
    @show_entries: (#dict) e.g. |{'stream': 'avg_frame_rate,height,width'}|
    """
    BINPATH = '/home/ubuntu/bin'
    BIN = 'ffprobe'
    OPT = Ffmpeg.OPT + [
        Opt('f', 'format'),
        Flag('unit'),
        Flag('prefix'),
        Flag('byte_binary_prefix'),
        Flag('sexagesimal'),
        Flag('pretty'),
        Opt('print_format', value='json'),
        Flag('sections'),
        Opt('select_streams'),
        Opt('show_data_hash'),
        Flag('show_error'),
        Flag('show_format'),
        Opt('show_format_entry'),
        Flag('show_packets'),
        Flag('show_frames'),
        Flag('show_streams'),
        Flag('show_chapters'),
        Flag('count_frames'),
        Flag('count_packets'),
        GroupOpt('read_interval', separator=','),
        Flag('private'),
        Flag('show_program_version'),
        Flag('show_library_versions'),
        Flag('show_versions'),
        Flag('show_pixel_formats'),
        KeyValOpt('show_entries', separator=':'),
        Flag('bitexact')
    ]

    # #
    # REMEMBER
    # ~~~~~~~~
    # ffprobe test_files/test/foo/1080.m4s -print_format json -show_entries stream -show_data_hash sha512
    # ffmpeg -i test_files/test/foo/360.m4s -f hash -hash sha512 -
    # #
    def _exec_pipe(self, *outputs, stdout=None, stderr=None):
        return super(Probe, self)._exec_pipe(
            *outputs,
            stdout=stdout,
            stderr=stderr,
            no_faststart=True
        )

    def is_playable(self, criteria=None):
        """ Tests if the file is considered playable based on given criteria
            compared against the JSON results of the FFPROBE command
        """
        #: TODO
        pass




'''
import json
from vital.debug import Timer
import transcend as ts
t = Timer()
t.start()
f = ts.Probe(show_format=True, show_streams=True)
f.input = ts.protocol.FileIn('test_files/test.vpx.webm')
# f.input = ts.protocol.S3In('s3://xfaps.queue/v/56/56294846-3904-45e1-b2af-c071cfbb629b.MOV')
# f.input = ts.protocol.S3In('s3://xfaps.queue/Sintel.2010.1080p.mkv')
# f.input = ts.protocol.S3In('s3://xfaps.queue/v/42/42e2bc9c-eb23-44b7-b870-1aba196b1172.mp4')
output = f.run()
print(json.dumps(json.loads(output.stdout.decode()), indent=2))
print(t.stop())
'''



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'MpegTs',


class MpegTs(Container):
    """
    MPEG transport stream muxer.
    This muxer implements ISO 13818-1 and part of ETSI EN 300 468.


    `MPEGTS container format`
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    @original_network_id: (#int) Set the original_network_id
        (default 0x0001). This is unique identifier of a network in DVB.
        Its main use is in the unique identification of a service through
        the path Original_Network_ID, Transport_Stream_ID.
    @transport_stream_id: (#int) Set the transport_stream_id
        (default 0x0001). This identifies a transponder in DVB.
    @service_id: (#int) Set the service_id (default 0x0001) also known
        as program in DVB.
    @service_type: (#str) Set the program service_type
        (default digital_tv), see below a list of pre defined values.
    @pmt_start_pid: (#int) Set the first PID for PMT (default 0x1000,
        max 0x1f00).
    @start_pid number: (#int) Set the first PID for data packets
        (default 0x0100, max 0x0f00).
    @m2ts_mode: (#bool) Enable m2ts mode if set to 1. Default value is
        -1 which disables m2ts mode.
    @mux_rate (muxrate): (#int) Set a constant muxrate (default VBR).
    @pcr_period: (#int) Override the default PCR retransmission time
        (default 20ms), ignored if variable muxrate is selected.
    @pat_period : (#int) Maximal time in seconds between PAT/PMT tables.
    @sdt_period: (#int) Maximal time in seconds between SDT tables.
    @pes_payload_size: (#int) Set minimum PES packet payload in bytes.
    @mpegts_flags: (flags)
    @copyts: (#bool) Preserve original timestamps, if value is set to 1.
        Default value is -1, which results in shifting timestamps so that they
        start from 0.
    @tables_version: (#int) Set PAT, PMT and SDT version (default 0, valid
        values are from 0 to 31, inclusively). This option allows updating
        stream structure so that standard consumer may detect the change.
    """
    OPT = [
        Opt('f', 'format', value='mpegts'),
        IntOpt('mpegts_original_network_id', 'original_network_id'),
        IntOpt('mpegts_transport_stream_id', 'transport_stream_id'),
        IntOpt('mpegts_service_id', 'service_id'),
        Opt('mpegts_service_type', 'service_type'),
        IntOpt('mpegts_pmt_start_pid', 'pmt_start_pid'),
        IntOpt('mpegts_start_pid', 'start_pid'),
        EnableOpt('mpegts_m2ts_mode', 'm2ts_mode', lower=-1),
        IntOpt('muxrate', 'mux_rate'),
        IntOpt('pcr_period'),
        IntOpt('pat_period'),
        ByteOpt('pes_payload_size'),
        FlagsOpt('mpegts_flags'),
        EnableOpt('mpegts_copyts', 'copyts', lower=-1),
        BoundedOpt('tables_version', lower=0, upper=31)
    ]



# -*- coding: utf-8 -*-
from ...cli import *
from .._protocol import Protocol


__all__ = 'IcecastOut',


class IcecastOut(Protocol):
    """
    `Icecast output protocol.`
    ~~~~~~~~~~~~~~~~~~~~~~~~
    Streams to icecast servers.

    @genre (ice_genre): (#str) Set the stream genre
    @name (ice_name): (#str) Set the stream name
    @description (ice_description): (#str) Set the stream description
    @url (ice_url): (#str) Set the stream website URL.
    @public (ice_public): (#bool) Set if the stream should be public.
        The default is 0 (not public).
    @user_agent: (#str) Override the User-Agent header. If not specified a
        string of the form "Lavf/<version>" will be used.
    @password: (#str) Set the Icecast mountpoint password.
    @content_type: (#str) Set the stream content type. This must be set if it
        is different from audio/mpeg.
    @legacy_icecast: (#bool) This enables support for Icecast versions < 2.4.0,
        that do not support the HTTP PUT method but the SOURCE method.

    """
    OPT = [
        Opt('ice_genre', 'genre'),
        Opt('ice_description', 'description'),
        Opt('ice_name', 'name'),
        Opt('ice_url', 'url'),
        EnableOpt('ice_public', 'public'),
        Opt('user_agent'),
        Opt('password'),
        Opt('content_type'),
        EnableOpt('legacy_icecast')
    ]

    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value='icecast://' + endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
"""
`Animated Gif`
==============
"""
from ..cli import *
from ._container import *


__all__ = 'Gif',


class Gif(Container):
    """
    `Gif container format`
    ~~~~~~~~~~~~~~~~~~~~~~
    @loop: (#int) et the number of times to loop the output. Use -1 for no
        loop, 0 for looping indefinitely (default).
    """
    OPT = [
        Opt('f', 'format', value='gif'),
        IntOpt('loop')
    ]



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Image2',


class Image2(Container):
    """
    The image file muxer writes video frames to image files.

    The output filenames are specified by a pattern, which can be used to
    produce sequentially numbered series of files. The pattern may contain the
    string "%d" or "%0Nd", this string specifies the position of the characters
    representing a numbering in the filenames. If the form "%0Nd" is used, the
    string representing the number in each filename is 0-padded to N digits.
    The literal character ’%’ can be specified in the pattern with the
    string "%%".

    If the pattern contains "%d" or "%0Nd", the first filename of the file
    list specified will contain the number 1, all the following numbers will
    be sequential.

    The pattern may contain a suffix which is used to automatically determine
    the format of the image files to write.

    For example the pattern "img-%03d.bmp" will specify a sequence of filenames
    of the form img-001.bmp, img-002.bmp, ..., img-010.bmp, etc. The pattern
    "img%%-%d.jpg" will specify a sequence of filenames of the form img%-1.jpg,
    img%-2.jpg, ..., img%-10.jpg, etc.


    `Image2 container format`
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    @start_number: (#int) Start the sequence from the specified number. Default
        value is 0.
    @update: (#bool) If set to 1, the filename will always be interpreted as
        just a filename, not a pattern, and the corresponding file will be
        continuously overwritten with new images. Default value is 0.
    @strftime: (#bool) If set to 1, expand the filename with date and time
        information from strftime(). Default value is 0.
    """
    OPT = [
        Opt('f', 'format', value='image2'),
        IntOpt('start_number'),
        EnableOpt('update'),
        EnableOpt('strftime')
    ]



# -*- coding: utf-8 -*-
"""
`VPX`
=====
https://trac.ffmpeg.org/wiki/Encode/VP8


Variable bitrates
~~~~~~~~~~~~~~~~~
- libvpx offers a variable bitrate mode by default. In this mode, it will
  simply try to reach the specified bit rate on average, e.g. 1 MBit/s.
  This is the "target bitrate".
- In addition to the "default" VBR mode, there's a constant quality mode
  (like in the x264 encoder) that will ensure that every frame gets the number
  of bits it deserves to achieve a certain quality level, rather than forcing
  the stream to have an average bit rate. This results in better overall
  quality and should be your method of choice when you encode video with
  libvpx. In this case, the target bitrate becomes the maximum allowed bitrate.
  You enable the constant quality mode with the CRF parameter.


Constant bitrates
~~~~~~~~~~~~~~~~~
Like most other encoders, libvpx offers a constant bitrate encoding mode as
well, which tries to encode the video in such a way that an average bitrate
is reached. This doesn't mean that every frame is encoded with the same amount
of bits (since it would harm quality), but the bitrate will be very constrained.
You should use constant bitrate encoding if you need your video files to have a
certain size, or if you're streaming the videos over a channel that only allows
a certain bit rate.
"""
from ...cli import *
from ...codec._codec import *


__all__ = 'Vpx', 'Vp8', 'Vp9',


class Vpx(VideoCodec):
    """
    VP8/VP9 format is supported through libvpx.

    To reduce the duplication of documentation, only the private options and
    some others requiring special attention are documented here. For the
    documentation of the undocumented generic options, see :class:VideoCodec.

    ``VPX-specific options``
    ~~~~~~~~~~~~~~~~~~~~~~~~
    Source: http://wiki.webmproject.org/ffmpeg

    @undershoot_pct (undershoot-pct): (#int 0-100) Set datarate undershoot
        (min) percentage of the target bitrate.
    @overshoot_pct (overshoot-pct): (#int 0-100) Set datarate overshoot (max)
        percentage of the target bitrate.
    @crf: (#int 4-63) constant rate factor
    @tune: (#str) 'psnr' or 'ssim'
    @quality: (#str) 'best', 'good', 'realtime'
    @speed (cpu-used): (#int -16-16) Set quality/speed ratio modifier.
        Higher values speed up the encode at the cost of quality.
    @noise_sensitivity (nr): (#int 0-6) Crude temporal noise filter.
    @static_thresh (static-thresh): (#int) Set a change threshold on blocks
        below which they will be skipped by the encoder.
    @slices: Note that FFmpeg’s slices option gives the total number of
        partitions, while vpxenc’s token-parts is given as log2(partitions).
    @max_intra_rate (max-intra-rate): (#int) Set maximum I-frame bitrate as a
        percentage of the target bitrate. A value of 0 means unlimited.
    @force_key_frames: (#bool) VPX_EFLAG_FORCE_KF
    @auto_alt_ref (auto-alt-ref): (#int) Enable use of alternate reference
        frames (2-pass only). 0 for disabled, 1 for enabled.
    @arnr_max_frames (arnr-max-frames): (#int) Set altref noise reduction max
        frame count.
    @arnr_type (arnr-type): (#str) 'backward', 'forward', 'centered'
    @arnr_strength (arnr-strength): (#int 0-6) Set altref noise reduction filter
        strength.
    @error_resilient (error-resilient): (#int) Enable error resiliency features.
        0 for disabled, 1 for enabled.
    """
    OPT = [
        Opt('c:v', 'codec', value='libvpx'),
        BoundedOpt('undershoot-pct', 'undershoot_pct',
                   lower=0, upper=100),
        BoundedOpt('overshoot-pct', 'overshoot_pct',
                   lower=0, upper=100),
        BoundedOpt('crf', lower=4, upper=63),
        OneOfOpt('tune', choices=('psnr', 'ssim')),
        OneOfOpt('quality', choices=('best', 'good', 'realtime')),
        BoundedOpt('cpu-used', 'speed', lower=-16, upper=16),
        BoundedOpt('nr', 'noise_sensitivity', lower=0, upper=6),
        IntOpt('static-thresh', 'static_thresh'),
        OneOfOpt('slices', cast=int, choices=(1, 2, 4, 8)),
        BoundedOpt('max-intra-rate', 'max_intra_rate',
                   lower=0, upper=100),
        Flag('force_key_frames'),
        EnableOpt('auto-alt-ref', 'auto_alt_ref'),
        IntOpt('arnr-max-frames', 'arnr_max_frames'),
        OneOfOpt('arnr-type', 'arnr_type',
                 choices=('backward', 'forward', 'centered')),
        BoundedOpt('arnr-strength', 'arnr_strength', lower=0, upper=6),
        EnableOpt('error-resilient', 'error_resilient')
    ]


Vp8 = Vpx


class Vp9(Vp8):
    """
    ``VP9-specific options``
    ~~~~~~~~~~~~~~~~~~~~~~~~
    @dash: (#int) Set MPEG-DASH segment duration.
    @lossless: (#bool) Enable lossless mode. 1 for enabled, 0 for disabled
    @tile_columns (tile-columns): (#int) Set number of tile columns to use.
        Note this is given as log2(tile_columns). For example, 8 tile columns
        would be requested by setting the tile-columns option to 3.
    @tile_rows (tile-rows): (#int) Set number of tile rows to use. Note this
        is given as log2(tile_rows). For example, 4 tile rows would be
        requested by setting the tile-rows option to 2.
    @frame_parallel (frame-parallel): (#bool) Enable frame parallel
        decodability features. 1 for enabled, 0 for disabled.
    @aq_mode (aq-mode): (#int 0-3) Set adaptive quantization mode
        (0: off (default), 1: variance 2: complexity, 3: cyclic refresh).
    @colorspace: (#str) 'rgb', 'bt709', 'unspecified', 'bt470bg', 'smpte170m',
        'smpte240m', 'bt2020_ncl'
    """
    OPT = Vpx.OPT + [
        Opt('c:v', 'codec', value='libvpx-vp9'),
        IntOpt('dash'),
        OneOfOpt('strict', choices=('very', 'strict', 'normal',
                                    'experimental'),
                           value='experimental'),
        IntOpt('tile-columns', 'tile_columns'),
        IntOpt('tile-rows', 'tile_rows'),
        EnableOpt('frame-parallel', 'frame_parallel'),
        BoundedOpt('aq-mode', 'aq_mode', lower=0, upper=3),
        OneOfOpt('colorspace',
                 choices=('rgb', 'bt709', 'unspecified', 'bt470bg',
                          'smpte170m', 'smpte240m', 'bt2020_ncl'))
    ]



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'TcpIn', 'TcpOut',


TCP_OPT = [
    EnableOpt('listen'),
    IntOpt('timeout'),
    IntOpt('listen_timeout'),
    ByteOpt('recv_buffer_size'),
    ByteOpt('send_buffer_size')
]


class _Tcp(Protocol):
    """
    `Transmission Control Protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @listen: (#bool) Listen for an incoming connection. Default value is 0.
    @timeout: (#int) Set raise error timeout, expressed in microseconds.
    @listen_timeout: (#int) Set listen timeout, expressed in milliseconds.
    @recv_buffer_size: (#str) Set receive buffer size, expressed bytes.
    @send_buffer_size: (#str) Set send buffer size, expressed bytes.
    """
    OPT = TCP_OPT


class TcpIn(_Tcp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class TcpOut(_Tcp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
import re
import os
import copy
import humanfriendly
from collections import OrderedDict
import subprocess


__all__ = (
    'Runnable',
    'OptSet',
    'Opt',
    'Flag',
    'FlagsOpt',
    'BitOpt',
    'ByteOpt',
    'TimeOpt',
    'IntOpt',
    'FloatOpt',
    'BoundedOpt',
    'EnableOpt',
    'KeyValOpt',
    'GroupOpt',
    'SizeOpt',
    'OneOfOpt'
)


class OptSet(object):
    __slots__ = '_opts', 'META'

    def __init__(self, *opt, **opts):
        self.META = dict()
        self._opts = OrderedDict()

        if len(opt):
            self.add(*opt)

        if len(opts):
            self.set(**opts)

    def __repr__(self):
        cmds = map(str, (x for opt in self.live_opts for x in opt))
        cmds = subprocess.list2cmdline(cmds)
        return '<%s: %s>' % (self.__class__.__name__, cmds)

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            if name == '_opts':
                raise
            if name not in self._opts:
                raise
            return self._opts[name]

    def __setattr__(self, name, value):
        try:
            if name != '_opts':
                self._opts[name].set_value(value)
            else:
                raise AttributeError
        except (AttributeError, KeyError):
            super(OptSet, self).__setattr__(name, value)

    def __delattr__(self, name):
        try:
            if name != '_opts':
                self._opts[name].set_value(None)
            else:
                raise AttributeError
        except (AttributeError, KeyError):
            super(OptSet, self).__delattr__(name)

    def __str__(self):
        return ' '.join(map(str, self.opts))

    def __iter__(self):
        for opt in self.opts:
            yield opt

    def add(self, *opt):
        for o in opt:
            if isinstance(o, (Opt, Flag)):
                _opt = self.get(o.name)
                if _opt is None:
                    self._opts[o.name] = copy.copy(o)
                else:
                    _opt.set_value(o.val)
                    _opt.opt = o.opt
            elif isinstance(o, (OptSet)):
                self.add(*o)
            else:
                raise TypeError('Invalid option: %s; Custom options must be '
                                'of type "transcend.Opt" "transcend.OptSet" or '
                                '"transcend.Flag".' % o.name)

    def get(self, opt_name, default=None):
        try:
            return self._opts[opt_name]
        except KeyError:
            return default

    def set(self, **opts):
        for name, value in opts.items():
            try:
                self._opts[name].set_value(value)
            except KeyError:
                raise NameError('Name %s is not a registered option.' % name)

    def list(self):
        for opt in sorted(self.opts, key=lambda x: x.name[0]):
            print(repr(opt))

    @property
    def opts(self):
        return [opt for opt in self._opts.values()]

    @property
    def opt_names(self):
        return [opt.name for opt in self._opts.values()]

    @property
    def live_opts(self):
        for opt in self._opts.values():
            if opt.val is not None:
                yield opt

    def items(self):
        return {name: value for name, value in self._opts.items()}

    def merge(self, *cls):
        for c in cls:
            try:
                self.add(*c._opts.values())
            except AttributeError:
                pass


class Runnable(OptSet):
    __slots__ = OptSet.__slots__

    def __repr__(self):
        return '<%s>' % ' '.join(self.get_cmd())

    def get_bin(self):
        return os.path.join(self.BINPATH or '', self.BIN)

    def get_cmd(self, no_bin=False):
        if not no_bin:
            yield self.get_bin()

        for opt in self.live_opts:
            for x in opt:
                yield str(x)


class Opt(object):
    __slots__ = ['name', 'opt', 'val', 'desc', 'cast', 'default']

    def __init__(self, opt, name=None, value=None, cast=str, default=None):
        self.name = name or opt
        self.opt = '-%s' % opt.lstrip('-') if len(opt) else ''
        self.cast = cast
        self.val = None
        self.default = default
        self.set_value(value)

    def __call__(self, *args, **kwargs):
        return self.set_value(*args, **kwargs)

    def __repr__(self):
        s = str(self)
        if len(s):
            return '<%s>' % s
        return '<%s: %s [%s]>' % (self.name, self.opt, self.cast)

    def __iter__(self):
        if len(self.opt):
            yield self.opt
        if self.val is None:
            yield str(self.default)
        else:
            yield str(self)

    def __str__(self):
        if self.val is not None:
            return str(self.val if self.val is not None else self.default)
        return ''

    def to_proc(self):
        if self.val:
            return list(self)
        return []

    def set_value(self, value):
        if value is None:
            self.val = None
        else:
            self.val = self.cast(value)
        return self


class IntOpt(Opt):
    __slots__ = Opt.__slots__

    def __init__(self, *args, **kwargs):
        super(IntOpt, self).__init__(*args, cast=int, **kwargs)


class FloatOpt(Opt):
    __slots__ = Opt.__slots__

    def __init__(self, *args, **kwargs):
        super(FloatOpt, self).__init__(*args, cast=float, **kwargs)


class BoundedOpt(Opt):
    __slots__ = Opt.__slots__ + ['_upper_bound', '_lower_bound']

    def __init__(self, opt, name=None, cast=int, lower=0, upper=100, **kwargs):
        self._lower_bound = lower
        self._upper_bound = upper
        super(BoundedOpt, self).__init__(opt, name=name, cast=cast, **kwargs)

    def set_value(self, value):
        if value is None:
            self.val = value
        else:
            value = self.cast(value)
            if self._lower_bound <= value <= self._upper_bound:
                self.val = value
            else:
                raise ValueError('Value `%s` not within bounds %s-%s.' %
                                 (value, self._lower_bound, self._upper_bound))
        return self


class EnableOpt(BoundedOpt):
    __slots__ = BoundedOpt.__slots__

    def __init__(self, opt, name=None, lower=0, upper=1, **kwargs):
        super(EnableOpt, self).__init__(
            opt, name=name, cast=int, upper=upper,  lower=lower, **kwargs)

    def set_value(self, value=True):
        if value is True:
            value = self._upper_bound
        if value is False:
            value = self._lower_bound
        return super(EnableOpt, self).set_value(value)


class Flag(Opt):
    __slots__ = Opt.__slots__

    def __repr__(self):
        s = str(self)
        if len(s):
            return '<%s>' % s
        return '<%s: %s [flag (not set)]>' % (self.name, self.opt)

    def __call__(self, value=True):
        return self.set_value(value)

    def __iter__(self):
        if self.val is None and self.default is True:
            yield str(self.opt)
        else:
            yield str(self.val)

    def __str__(self):
        if self.val is not None:
            return str(self.val)
        return ''

    def to_proc(self):
        if self.val:
            return list(iter(self))
        return []

    def set_value(self, value):
        if value in {True, False, None}:
            self.val = self.opt if value is True else None
        return self


class FlagsOpt(Opt):
    __slots__ = Opt.__slots__

    def set_value(self, flag=None, *flags):
        if flag is None:
            self.val = None
        else:
            flags = list(flags)
            flags.insert(0, flag)
            self.val = ' '.join(map(lambda x: '+' + str(x).strip('+'), flags))
        return self


class BitOpt(Opt):
    __slots__ = Opt.__slots__
    _RE_SIZE = re.compile(r"""([\d.]+)([^\d.])""")

    def set_value(self, bits):
        if isinstance(bits, str):
            bits = bits.strip()
            bitsearch = self._RE_SIZE.search(bits)
            if bitsearch is not None:
                bits = float(bitsearch.group(1))
                abbr = bitsearch.group(2).strip().lower()
            else:
                abbr = None
                bits = float(bits)
            if abbr is None:
                self.val = int(bits)
            elif abbr == 'k':
                self.val = int(bits * 1000)
            elif abbr == 'm':
                self.val = int(bits * (1000 ** 2))
            elif abbr == 'g':
                self.val = int(bits * (1000 ** 3))
        else:
            if bits is None:
                self.val = bits
            else:
                self.val = int(bits)
        return self


class ByteOpt(Opt):
    __slots__ = Opt.__slots__
    _RE_SIZE = re.compile(r"""([\d.]+)([^\d.])""")

    def set_value(self, bytes):
        if isinstance(bytes, str):
            self.val = int(humanfriendly.parse_size(bytes))
        else:
            if bytes is None:
                self.val = bytes
            else:
                self.val = int(bytes)
        return self


class TimeOpt(Opt):
    __slots__ = Opt.__slots__

    def set_value(self, seconds):
        if seconds is None:
            self.val = None
        else:
            try:
                self.val = float(seconds)
            except (ValueError, TypeError):
                time = seconds.split(':')
                seconds = float(time[-1].lstrip('0') or 0)
                minutes = int(time[-2].lstrip('0') or 0)
                try:
                    hours = int(time[-3].lstrip('0') or 0)
                except:
                    hours = 0
                self.val = float(seconds + (minutes * 60) + (hours * 60 * 60))
        return self


class SizeOpt(Opt):
    __slots__ = Opt.__slots__

    def set_value(self, size):
        """ @size: (#tuple) """
        if size is None:
            self.val = None
        elif not isinstance(size, str):
            self.val = 'x'.join(map(str, size))
        else:
            self.val = 'x'.join(map(str.strip, size.lower().split('x')))
        return self


class KeyValOpt(Opt):
    __slots__ = Opt.__slots__ + ['separator']

    def __init__(self, opt, name=None, separator=',', **kwargs):
        self.separator = separator
        super(KeyValOpt, self).__init__(opt, name, **kwargs)

    def _to_str(self, pairs):
        if not isinstance(pairs, dict):
            return str(pairs)
        out = []
        add_out = out.append
        for name, value in pairs.items():
            add_out('%s=%s' % (name, value))
        return self.separator.join(out)

    def __str__(self):
        if self.val is not None:
            return self._to_str(self.val)
        return ''

    def set_value(self, pairs):
        """ @pairs: (#dict) """
        if pairs is None:
            self.val = None
        # elif isinstance(pairs, str) or pairs is None:
        #     self.val = pairs
        else:
            self.val = pairs
        return self

    def set(self, **opt):
        return self.update(opt)

    def update(self, opt):
        old_val = self.val.copy()
        old_val.update(opt)
        self.set_value(old_val)
        return self


class GroupOpt(KeyValOpt):
    __slots__ = KeyValOpt.__slots__

    def set_value(self, items):
        """ @pairs: (#dict) """
        if isinstance(items, str) or items is None:
            self.val = items
        else:
            self.val = self.separator.join(map(lambda x: str(self.cast(x)),
                                               items))
        return self


class OneOfOpt(Opt):
    __slots__ = Opt.__slots__ + ['choices']

    def __init__(self, opt, name=None, cast=str, choices=None, **kwargs):
        self.choices = tuple(choices or [])
        super(OneOfOpt, self).__init__(opt, name=name, cast=cast, **kwargs)

    def set_value(self, val):
        if val is None:
            self.val = val
        elif self.cast(val) in self.choices:
            self.val = self.cast(val)
        else:
            raise ValueError('`%s` not found in %s. Must be one of: \n  %s' %
                             (val, self.name, tuple(self.choices)))
        return self



# -*- coding: utf-8 -*-
from pathlib import Path
from ..cli import *
from ._protocol import Protocol


__all__ = 'FileIn', 'FileOut',


class _File(Protocol):
    OPT = [EnableOpt('truncate'), ByteOpt('blocksize')]


class FileIn(_File):
    def __init__(self, filename, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'filename', value=filename),
            **opts
        )


class FileOut(_File):
    def __init__(self, filename, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'filename', value=filename),
            **opts
        )

        try:
            self.path = Path(filename)
            self.path.parent.mkdir(0o744, parents=True)
        except FileExistsError:
            pass



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'TlsIn', 'TlsOut',


TLS_OPT = [
    Opt('ca_file'),
    Opt('cert_file'),
    Opt('key_file'),
    EnableOpt('tls_verify', 'verify'),
    EnableOpt('listen')
]


class _Tls(Protocol):
    """
    `Transport Layer Security (TLS) protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @ca_file: (#str) A file containing certificate authority (CA) root
        certificates to treat as trusted. If the linked TLS library contains
        a default this might not need to be specified for verification to work,
        but not all libraries and setups have defaults built in. The file must
        be in OpenSSL PEM format.
    @cert_file: (#str) A file containing a certificate to use in the handshake
        with the peer. (When operating as server, in listen mode, this is more
        often required by the peer, while client certificates only are mandated
        in certain setups.)
    @key_file: (#str) A file containing the private key for the certificate.
    @verify (tls_verify): (#bool) If enabled, try to verify the peer that we
        are communicating with. Note, if using OpenSSL, this currently only
        makes sure that the peer certificate is signed by one of the root
        certificates in the CA database, but it does not validate that the
        certificate actually matches the host name we are trying to connect
        to. (With GnuTLS, the host name is validated as well.)
    @listen: (#bool) If enabled, listen for connections on the provided port,
        and assume the server role in the handshake instead of the client role.
    """
    OPT = TLS_OPT


class TlsIn(_Tls):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class TlsOut(_Tls):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



#: See: http://www.ffmpeg.org/ffmpeg-protocols.html#tls
from ._protocol import Protocol
from .input import *
from .output import *
from .file import *
from .ftp import *
from .http import *
from .pipe import *
from .rtmp import *
from .rtp import *
from .s3 import *
from .sftp import *
from .srtp import *
from .tcp import *
from .tls import *
from .udp import *
from .unix import *
from . import input, output



# -*- coding: utf-8 -*-
"""
`FLAC`
============
libflac wrapper for FFMPEG
"""
from ...cli import *
from .._codec import *


__all__ = 'Flac',


class Flac(AudioCodec):
    """
    @compression_level: (#int) Sets the compression level, which chooses
        defaults for many other options if they are not set explicitly.
    @frame_size: (#int) Sets the size of the frames in samples per channel.
    @lpc_coeff_precision: (#int) Sets the LPC coefficient precision, valid
        values are from 1 to 15, 15 is the default.
    @lpc_type: (#str) Sets the first stage LPC algorithm. 'none', 'fixed',
        'levinson', 'cholesky'
    @lpc_passes: (#int) Number of passes to use for Cholesky factorization
        during LPC analysis
    @min_partition_order: (#int) The minimum partition order
    @max_partition_order: (#int) The maximum partition order
    @prediction_order_method: (#str) 'estimation', '2level', '4level', '8level',
        'search', 'log'
    @channel_mode (ch_mode): (#str) Sets channel mode; 'auto', 'indep',
        'left_side', 'right_side', 'mid_side'
    @exact_rice_parameters: (#bool) Chooses if rice parameters are calculated
        exactly or approximately. if set to 1 then they are chosen exactly,
        which slows the code down slightly and improves compression slightly.
    @multi_dim_quant: (#bool) Multi Dimensional Quantization. If set to 1 then
        a 2nd stage LPC algorithm is applied after the first stage to finetune
        the coefficients. This is quite slow and slightly improves compression.
    """
    OPT = [
        Opt('c:a', 'codec', value='libflac'),
        IntOpt('compression_level'),
        IntOpt('frame_size'),
        BoundedOpt('lpc_coeff_precision', lower=1, upper=15),
        OneOfOpt('lpc_type',
                 choices=('none', 'fixed', 'levinson', 'cholesky')),
        IntOpt('lpc_passes'),
        IntOpt('min_partition_order'),
        IntOpt('max_partition_order'),
        OneOfOpt('prediction_order_method',
                 choices=('estimation', '2level', '4level', '8level',
                          'search', 'log')),
        OneOfOpt('ch_mode', 'channel_mode',
                 choices=('auto', 'indep', 'left_side', 'right_side',
                          'mid_side')),
        EnableOpt('exact_rice_parameters'),
        EnableOpt('multi_dim_quant')
    ]



from ...codec import H264, MotionEstimation, BFrames, Aac, Quantizer
from ...output import Video, Audio
from ...format import Mp4
from ...utils import calc_scale
from ..audio.aac import LoFiAudio, MidFiAudio, HiFiAudio


class DefaultMe(MotionEstimation):
    def __init__(self, input=None, **opts):
        super(DefaultMe, self).__init__(
            me_method='umh',
            me_range=32,
            subq_est=10,
        )


class LowQuantizer(Quantizer):
    def __init__(self, input=None, **opts):
        super(LowQuantizer, self).__init__(
            min=3,
            curve_compression=0.9
        )


class DefaultQuantizer(Quantizer):
    def __init__(self, input=None, **opts):
        super(DefaultQuantizer, self).__init__(
            curve_compression=0.9
        )


class LowBFrames(BFrames):

    def __init__(self, input=None, **opts):
        super(LowBFrames, self).__init__(
            strategy=0,
            max=0,
        )


class DefaultBFrames(BFrames):

    def __init__(self, input=None, **opts):
        super(DefaultBFrames, self).__init__(
            strategy=1,
            max=16,
        )


class HighBFrames(BFrames):

    def __init__(self, input=None, **opts):
        super(HighBFrames, self).__init__(
            strategy=2,
            max=3,
        )


class BaseH264(Video):

    def __init__(self, *opt, **opts):
        super().__init__(
            DefaultMe(),
            DefaultQuantizer(),
            *opt,
            pixel_format='yuv420p',
            codec=H264(
                # key_interval=120,
                # min_key_interval=1,
                frame_rate=30,
                fast_pskip=True,
                scene_change_thresh=40,
            ),
            format=Mp4(),
            **opts
        )

        if self.output is not None and self.codec.size.val is not None:
            self.output.filters.set(scale=calc_scale(self.codec.size.val))



class AnyDevice(BaseH264):

    def __init__(self, input=None, **opts):
        super().__init__(
            LowBFrames(),
            LowQuantizer(),
            audio=LoFiAudio,
            **opts
        )
        self.codec.set(
            size='424x240',
            max_bitrate='576k',
            buffer_size='864k',
            frame_rate=24,
            profile='baseline',
            level=3,
            partitions=['p4x4'],
            trellis=0
        )


class AnyDevice360p(AnyDevice):

    def __init__(self, input=None, **opts):
        super().__init__(**opts)
        self.codec.set(size='640x360',
                       level=3,
                       frame_rate=30,
                       max_bitrate='1024k',
                       buffer_size='1536k')


class Main360p(BaseH264):

    def __init__(self, input=None, **opts):
        super().__init__(
            DefaultBFrames(),
            audio=LoFiAudio,
            **opts)
        self.codec.set(
            size='640x360',
            max_bitrate='1024k',
            buffer_size='1536k',
            profile='main',
            level=3.1,
            coder=1,
            trellis=2
        )


class High360p(Main360p):

    def __init__(self, input=None, **opts):
        super().__init__(**opts)
        self.codec.set(profile='high')


class Main480p(BaseH264):

    def __init__(self, input=None, **opts):
        super().__init__(
            DefaultBFrames(),
            audio=MidFiAudio,
            **opts)
        self.codec.set(
            size='848x480',
            max_bitrate='1536k',
            buffer_size='2304k',
            profile='main',
            level=3.1,
            coder=1,
            trellis=2,
            rc_lookahead=32
        )


class High480p(Main480p):

    def __init__(self, input=None, **opts):
        super().__init__(**opts)
        self.codec.set(profile='high')


class Main576p(BaseH264):
    def __init__(self, input=None, **opts):
        super().__init__(
            DefaultBFrames(),
            audio=MidFiAudio,
            **opts)
        self.codec.set(
            size='960x576',
            max_bitrate='2176k',
            buffer_size='3264k',
            profile='main',
            level=3.1,
            coder=1,
            trellis=2,
            rc_lookahead=40
        )


class High576p(Main576p):
    def __init__(self, input=None, **opts):
        super().__init__(**opts)
        self.codec.set(profile='high')


class High720p(BaseH264):
    def __init__(self, input=None, **opts):
        super().__init__(
            DefaultBFrames(),
            audio=HiFiAudio,
            **opts)
        self.codec.set(
            size='1280x720',
            max_bitrate='5242k',
            buffer_size='7863k',
            profile='high',
            level=4,
            coder=1,
            trellis=2,
            rc_lookahead=60
        )


class High1080p(BaseH264):
    def __init__(self, input=None, **opts):
        super().__init__(
            DefaultBFrames(),
            audio=HiFiAudio,
            **opts
        )
        self.codec.set(
            size='1920x1080',
            max_bitrate='7552k',
            buffer_size='11328k',
            frame_rate=60,
            profile='high',
            level=4.1,
            coder=1,
            trellis=2,
            rc_lookahead=60
        )


class Super1080p(BaseH264):
    def __init__(self, input=None, **opts):
        super().__init__(
            HighBFrames(),
            audio=HiFiAudio,
            **opts)
        self.codec.set(
            size='1920x1080',
            max_bitrate='12M',
            buffer_size='18M',
            frame_rate=60,
            profile='high',
            level=4.1,
            fast_pskip=False,
            deblock_alpha=-2,
            deblock_beta=-1,
            coder=1,
            trellis=2,
            rc_lookahead=120
        )


class High1440p(BaseH264):
    def __init__(self, input=None, **opts):
        super().__init__(
            HighBFrames(),
            audio=HiFiAudio,
            **opts)
        self.codec.set(
            size='2560x1440',
            max_bitrate='18M',
            buffer_size='30.48M',
            frame_rate=60,
            profile='high',
            level=4.1,
            fast_pskip=False,
            deblock_alpha=-2,
            deblock_beta=-1,
            coder=1,
            trellis=2,
            rc_lookahead=175
        )

High2k = High1440p


class High2160p(BaseH264):
    def __init__(self, input=None, **opts):
        super().__init__(
            HighBFrames(),
            audio=HiFiAudio,
            **opts)
        self.codec.set(
            size='3840×2160',
            max_bitrate='55M',
            buffer_size='82.5M',
            frame_rate=60,
            profile='high',
            level=4.1,
            fast_pskip=False,
            deblock_alpha=-2,
            deblock_beta=-1,
            coder=1,
            trellis=2,
            rc_lookahead=175
        )

High4k = High2160p



# -*- coding: utf-8 -*-
import sys
import time
import subprocess
import urllib
from collections import namedtuple
from threading import Thread
try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty

from vital.debug import logg

from .cli import *
from .exceptions import FastStartError
from . import faststart


__all__ = 'Ffmpeg',


FfmpegResult = namedtuple('FfmpegResult', 'stdout stderr')


class Ffmpeg(Runnable):
    '''
    -loglevel loglevel  set logging level
    -v loglevel         set logging level
    -report             generate a report
    -max_alloc bytes    set maximum size of a single allocated block
    -y                  overwrite output files
    -n                  never overwrite output files
    -ignore_unknown     Ignore unknown stream types
    -stats              print progress report during encoding
    -max_error_rate ratio of errors (0.0: no errors, 1.0 max errors)
    -bits_per_raw_sample number  set the number of bits per raw sample
    -vol volume         change audio volume (256=normal)
    '''
    __slots__ = 'input', '_opts', '_with_schedtool'
    BINPATH = '/home/ubuntu/bin'
    BIN = 'ffmpeg'
    OPT = [
        Flag('y', 'overwrite'),
        Flag('n', 'never_overwrite'),
        Flag('h', 'help'),
        Flag('ignore_unknown'),
        Flag('stats', 'show_stats'),
        Flag('generate_report', 'report'),
        ByteOpt('max_alloc'),
        BoundedOpt(
            'max_error_rate', lower=0.0, upper=1.0, cast=float
        ),
        IntOpt('bits_per_raw_sample', 'bits_per_sample'),
        IntOpt('vol', 'volume'),
        OneOfOpt(
            'loglevel',
            choices=(
                "quiet",
                "panic",
                "fatal",
                "error",
                "warning",
                "info",
                "verbose",
                "debug",
                "trace"
            )
        )
    ]

    def __init__(self, input=None, with_schedtool=True, *opt, **opts):
        super(Ffmpeg, self).__init__(*(list(opt) + self.OPT), **opts)
        self._with_schedtool = with_schedtool
        self.open(input)

    def open(self, input):
        if input is None:
            pass
        elif input.startswith('s3://'):
            pass
        elif input.startswith('ftp://'):
            pass
        elif input.startswith('sftp://'):
            pass
        elif input.startswith('http://'):
            pass
        else:
            input = humanfriendly.parse_path(input)
        self.input = input

    def get_input(self):
        if self.input is not None:
            for opt in self.input.live_opts:
                for x in opt.to_proc():
                    yield x

    SCHED_CMD = ['sudo', 'schedtool', '-R', '-p', '99', '-e']

    def get_cmd(self, *outputs):
        if self._with_schedtool:
            for cmd in self.SCHED_CMD:
                yield cmd
        for cmd in super(Ffmpeg, self).get_cmd():
            yield cmd
        for cmd in self.get_input():
            yield cmd
        for output in outputs:
            for opt in output.live_opts:
                for step in opt:
                    yield str(step)

    def get_pretty_cmd(self, *outputs):
        if self._with_schedtool:
            for cmd in self.SCHED_CMD:
                yield cmd
        for cmd in super(Ffmpeg, self).get_cmd():
            yield cmd
        for cmd in self.get_input():
            yield cmd
        for output in outputs:
            yield '\\\n'
            for opt in output.live_opts:
                for step in opt:
                    yield str(step)

    def _process_stdout(self, stdout_queue, stdout, STDOUT):
        try:
            std = stdout_queue.get_nowait()
            if stdout is not None:
                std = stdout(std)
            if std is not None:
                STDOUT += std
        except Empty:
            pass
        return STDOUT

    @staticmethod
    def _handle_stdout(stdout, queue):
        ''' bytes from the stdout pipe pass through here '''
        for line in iter(stdout.readline, b''):
            # line == line of bytes from stdout
            queue.put_nowait(line)
        stdout.close()

    def handle_stdout(self, stdout):
        """ @stdout: listens to stdout pipe
            -> :class:Queue
        """
        q = Queue()
        t = Thread(target=self._handle_stdout, args=(stdout, q))
        t.daemon = True
        t.start()
        return q

    def _get_stream_iterator(self, proc, no_faststart):
        # return self.input.iter_body()
        if not no_faststart:
            try:
                fs = faststart.FastStart()
                stream = fs.from_stream(self.input.body)
                proc.stdin.write(next(stream))
            except FastStartError as e:
                def _stream(fs):
                    fs.cache.seek(0)
                    yield fs.cache.read()
                    for data in fs.iter_stream():
                        yield data
                stream = _stream(fs)
        else:
            stream = self.input.iter_body()

        return stream

    def _exec_pipe(self, *outputs, stdout=None, stderr=None,
                   no_faststart=False):
        builtin_modules = sys.builtin_module_names
        proc = subprocess.Popen(self.get_cmd(*outputs),
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                close_fds='posix' in builtin_modules)

        stdout_queue = self.handle_stdout(proc.stdout)
        stderr_queue = self.handle_stdout(proc.stderr)
        STDERR = b''
        STDOUT = b''
        stream = self._get_stream_iterator(proc, no_faststart)

        while proc.poll() is None:
            if not proc.stdin.closed:
                try:
                    proc.stdin.write(next(stream))
                except StopIteration:
                    proc.stdin.close()
                except BrokenPipeError:
                    # NEEDS A HANDLER
                    proc.stdin.close()
                except:
                    continue

            # ts.protocol.Pipe outputs to stdout
            # mp4 muxer does not support non seekable output
            STDOUT = self._process_stdout(stdout_queue, stdout, STDOUT)
            STDERR = self._process_stdout(stderr_queue, stderr, STDERR)

        #: Finishes off the thread queues
        time.sleep(1.0)
        while not stdout_queue.empty():
            STDOUT = self._process_stdout(stdout_queue, stdout, STDOUT)
        while not stderr_queue.empty():
            STDERR = self._process_stdout(stderr_queue, stderr, STDERR)

        return FfmpegResult(STDOUT, STDERR)

    def _exec_native(self, *outputs, stdout=None, stderr=None):
        builtin_modules = sys.builtin_module_names
        proc = subprocess.Popen(
            self.get_cmd(*outputs),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds='posix' in builtin_modules
        )

        STDOUT, STDERR = proc.communicate()

        return FfmpegResult(STDOUT, STDERR)

    def save(self, *outputs, stdout=None, stderr=None):
        logg('\n',
             subprocess.list2cmdline(self.get_pretty_cmd(*outputs))
        ).log('Executing')

        if hasattr(self.input, 'body'):
            result = self._exec_pipe(*outputs, stdout=stdout, stderr=stderr)
        else:
            result = self._exec_native(*outputs, stdout=stdout, stderr=stderr)

        try:
            self.input.close()
        except AttributeError:
            pass

        return result

    run = save



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Hash', 'Md5', 'FrameHash', 'FrameMd5',


class Hash(Container):
    """
    ``Hash container format``
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    @hash: (#str) Use the cryptographic hash function specified by the
        string algorithm: 'MD5', 'murmur3', 'RIPEMD128', 'RIPEMD160',
        'RIPEMD256', 'RIPEMD320', 'SHA160', 'SHA224', 'SHA256', 'SHA512/224',
        'SHA512/256', 'SHA384', 'SHA512', 'CRC32', 'adler32'
    """
    OPT = [
        Opt('f', 'format', value='hash'),
        OneOfOpt(
            'hash',
            choices=(
                'MD5',
                'murmur3',
                'RIPEMD128',
                'RIPEMD160',
                'RIPEMD256',
                'RIPEMD320',
                'SHA160',
                'SHA224',
                'SHA256',
                'SHA512/224',
                'SHA512/256',
                'SHA384',
                'SHA512',
                'CRC32',
                'adler32'
            )
        )
    ]


class Md5(Container):
    """
    ``MD5 container format``
    """
    OPT = [Opt('f', 'format', value='md5')]


class FrameMd5(Container):
    """
    ``FrameMD5 container format``
    """
    OPT = [Opt('f', 'format', value='framemd5')]


class FrameHash(Hash):
    """
    ``FrameHash container format``
    """
    OPT = [Opt('f', 'format', value='framehash')]



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Nut',


class Nut(Container):
    """
    `Nut container format`
    ~~~~~~~~~~~~~~~~~~~~~~
    @syncpoints: (flags)
    @write_index: (#bool) Write index at the end, the default is to write an
        index.
    """
    OPT = [
        Opt('f', 'format', value='nut'),
        FlagsOpt('syncpoints'),
        EnableOpt('write_index')
    ]



# -*- coding: utf-8 -*-
from ._container import Container
from .aac import *
from .aiff import *
from .asf import *
from .chromaprint import *
from .gif import *
from .hashing import *
from .image2 import *
from .matroska import *
from .mp3 import *
from .mp4 import *
from .mpegts import *
from .null import *
from .nut import *
from .ogg import *
from .streaming import *
from . import streaming



import os
import math
from ....codec import H264, Vpx, Vp9, MotionEstimation, BFrames,\
                                   Aac, Opus
from ....output import Video, Audio
from ....format import Hls, Mp4
from .. import h264
# from transcend.coding.utils import calc_gop


def calc_gop(frame_rate, hls_time=3):
    # https://github.com/vbence/stream-m#fragments
    return frame_rate * hls_time


#: HLS audio presets
AnyHlsAudio = Aac(bitrate='64k', sample_rate='44100', channels=2)
LoFiHlsAudio = Aac(bitrate='96k', sample_rate='44100', channels=2)
MidFiHlsAudio = Aac(bitrate='160k', sample_rate='44100', channels=2)
HiFiHlsAudio = Aac(bitrate='256k', sample_rate='44100', channels=6)


class BaseHlsAacAudio(Audio):
    def set_hls_opts(self):
        # FRAG SIZE
        # https://github.com/vbence/stream-m#fragments
        if not hasattr(self.format, 'hls_flags'):
            self.format.merge(Hls())

        self.format.movflags('frag_keyframe', 'empty_moov')
        # self.format.movflags('empty_moov')
        # self.format.hls_flags('single_file', 'append_list', 'round_durations')
        self.format.hls_flags('append_list')
        self.format.playlist_type('vod')
        self.format.allow_cache(True)
        self.format.list_size(0)

        bitrate = self.codec.bitrate.val or\
                  self.codec.max_bitrate.val or\
                  self.codec.min_bitrate.val

        if not self.format.time.val:
            self.format.time(3)

        try:
            fn, fe = os.path.splitext(self.output.path.name)
            op = self.output.path.parent.joinpath(fn)
            # self.format.init_filename(f'{op.parts[-1]}/init.m4a')
            self.format.init_filename(op.joinpath('init.m4a'))
            self.format.segment_filename(op.joinpath('%d.m4s'))
            op.mkdir(mode=0o777, parents=True, exist_ok=True)
        except AttributeError:
            pass

        if not self.format.segment_type.val:
            self.format.segment_type('fmp4')

        self.disable_video()
        self.disable_subtitles()


class AacxAnyDevice(BaseHlsAacAudio):

    def __init__(self, *opt, **opts):
        super().__init__(
            *opt,
            codec=AnyHlsAudio,
            format=Mp4(),
            **opts
        )
        self.set_hls_opts()


class AacxLoFi(BaseHlsAacAudio):

    def __init__(self, *opt, **opts):
        super().__init__(
            *opt,
            codec=LoFiHlsAudio,
            format=Mp4(),
            **opts
        )
        self.set_hls_opts()


class AacxMidFi(BaseHlsAacAudio):

    def __init__(self, *opt, **opts):
        super().__init__(
            *opt,
            codec=MidFiHlsAudio,
            format=Mp4(),
            **opts
        )
        self.set_hls_opts()


class AacxHiFi(BaseHlsAacAudio):

    def __init__(self, *opt, **opts):
        super().__init__(
            *opt,
            codec=HiFiHlsAudio,
            format=Mp4(),
            **opts
        )
        self.set_hls_opts()


#: Hls
#  https://kvssoft.wordpress.com/2015/01/28/mpeg-dash-gop/
class BaseHlsH264(h264.BaseH264):

    def set_hls_opts(self):
        self.codec.scene_change_thresh(0)
        self.format.movflags('frag_keyframe', 'empty_moov')
        self.format.movflags('empty_moov')

        if not hasattr(self.format, 'hls_flags'):
            self.format.merge(Hls())

        # FRAG SIZE
        # https://github.com/vbence/stream-m#fragments
        if not self.format.time.val:
            self.format.time(3)

        key_int = calc_gop(self.codec.frame_rate.val or 30, self.format.time.val)

        # if self.codec.max_bitrate.val and not self.codec.bitrate.val:
        #     self.codec.bitrate(self.codec.max_bitrate.val)
        self.codec.key_interval(key_int)
        self.codec.flags('cgop')
        self.codec.min_key_interval(key_int)

        self.format.list_size(0)
        # self.format.hls_flags('single_file', 'append_list', 'round_durations')
        self.format.hls_flags('append_list')
        self.format.playlist_type('vod')
        self.format.allow_cache(True)
        try:
            fn, fe = os.path.splitext(self.output.path.name)
            op = self.output.path.parent.joinpath(fn)
            #self.format.init_filename(f'{op.parts[-1]}/init.m4v')
            self.format.init_filename(op.joinpath('init.m4v'))
            self.format.segment_filename(op.joinpath('%d.m4s'))
            op.mkdir(mode=0o777, parents=True, exist_ok=True)
        except AttributeError:
            pass

        if not self.format.segment_type.val:
            self.format.segment_type('fmp4')

        self.audio = None
        self.disable_audio()
        self.disable_subtitles()


class H264xAnyDevice(h264.AnyDevice, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.codec.set(
            profile='main',
            level=3.1,
            trellis=2
        )
        self.set_hls_opts()


class H264xAnyDevice360p(h264.AnyDevice360p, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.set_hls_opts()


class H264x360p(h264.Main360p, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.set_hls_opts()


class H264x480p(h264.High480p, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.set_hls_opts()


class H264x576p(h264.High576p, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.set_hls_opts()


class H264x720p(h264.High720p, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.set_hls_opts()


class H264x1080p(h264.High1080p, BaseHlsH264):

    def __init__(self, *opt, **opts):
        super().__init__(*opt, **opts)
        self.set_hls_opts()


'''
FRAGMENTS

Live streams consist of fragments (self-contained units of frames which are not referencing any frame outside the fragment). A fragment always starts with a key-frame (intra-frame) therefore it is important that the encoder put in key frames regularly as this determines the size of the fragments.

The ideal fragment size is around 200 kBytes (or 1600 kbits). The key frame interval can be calculated with this formula:

1600k / <bitrate> * <framerate>
e.g. if you are publishing a 500 kbit stream with 16 fps, then: 1600 / 500 * 16 = 51.2 (or 1600000 / 500000 * 16 = 51.2) so every 52nd video frame should be a key frame.

The server splits fragments when it seems necessary. A soft minimum for frame size is currently 100k (no new fragment is started if a new key frame arrives within 100 kBytes from a previous key frame).

The hard maximum for a fragments is 2048 kBytes. This is twice the size needed to hold 2 seconds of a 4096 kbit/sec HD stream.
'''



# -*- coding: utf-8 -*-
"""
`MP3`
============
libmp3lame wrapper for FFMPEG
"""
from ...cli import *
from .._codec import *


__all__ = 'Mp3',


class Mp3(AudioCodec):
    """
    @compression_level: (#int) Set algorithm quality. Valid arguments are
        integers in the 0-9 range, with 0 meaning highest quality but slowest,
        and 9 meaning fastest while producing the worst quality.
    @reservoir: (#bool) Enable use of bit reservoir when set to 1. Default
        value is 1. LAME has this enabled by default.
    @joint_stereo: (#bool) Enable the encoder to use (on a frame by frame basis)
        either L/R stereo or mid/side stereo. Default value is 1.
    @abr: (#bool) Enable the encoder to use ABR when set to 1. The lame --abr
        sets the target bitrate, while this options only tells FFmpeg to use
        ABR still relies on b to set bitrate.
    """
    OPT = [
        Opt('c:a', 'codec', value='libmp3lame'),
        BoundedOpt('compression_level', lower=0, upper=9),
        EnableOpt('reservoir'),
        EnableOpt('joint_stereo'),
        EnableOpt('abr')
    ]



# -*- coding: utf-8 -*-
from ..cli import *


__all__ = 'Protocol',


DEFAULT_PROTOCOL_OPT = [
    GroupOpt('protocol_whitelist', 'whitelist', separator=','),
    IntOpt('rw_timeout', 'rw_timeout')
]


class Protocol(OptSet):
    """
    The libavformat library provides some generic global options, which can be
    set on all the protocols. In addition each protocol may support so-called
    private options, which are specific for that component.

    `Protocol`
    ~~~~~~~~~~
    @whitelist (protocol_whitelist): (#tuple(#str)) Set a list of allowed
        protocols.
    @rw_timeout (rw_timeout): (#int) Maximum time to wait for (network) read/write
        operations to complete, in microseconds.
    """
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_PROTOCOL_OPT + self.OPT + list(opt)), **opts)

    def __call__(self, *args, **kwargs):
        return self.__class__(*args, **kwargs)



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'UnixIn', 'UnixOut',


UNIX_OPT = [IntOpt('timeout'), EnableOpt('listen')]


class _Unix(Protocol):
    """
    `Unix domain sockets`
    ~~~~~~~~~~~~~~~~~~~~~
    @timeout: (#int) Timeout in ms.
    @listen: (#bool) Create the Unix socket in listening mode.
    """
    OPT = UNIX_OPT


class UnixIn(_Unix):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class UnixOut(_Unix):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



from multiprocessing import cpu_count
from ...codec import Vpx, Vp9, Quantizer
from ...output import Video, Audio
from ...format import WebM
from ..audio import opus, vorbis


class BaseVpx(Video):
    def __init__(self, input=None, **opts):
        super(BaseVpx, self).__init__(
            Quantizer(min=4, max=63),
            pixel_format='yuv420p',
            codec=Vpx(threads=cpu_count(),
                      crf=10),
            format=format.WebM(),
            **opts)


class BaseVp9(Video):
    def __init__(self, input=None, **opts):
        super(BaseVp9, self).__init__(
            Quantizer(min=4, max=63),
            pixel_format='yuv420p',
            codec=Vpx(threads=cpu_count(),
                      key_interval=120,
                      crf=10),
            format=WebM(),
            **opts)


class AnyDevice(BaseVpx):

    def __init__(self, input=None, **opts):
        super(AnyDevice, self).__init__(
            audio=vorbis.LoFiAudio,
            **opts)
        self.codec.set(size='424x240',
                       bitrate='712k',
                       max_bitrate='712k',
                       buffer_size='1068k',
                       quality='realtime',
                       speed=9,
                       crf=12)


class AnyDevice360p(AnyDevice):

    def __init__(self, input=None, **opts):
        super(AnyDevice360p, self).__init__(
            **opts)
        self.codec.set(size='640x360',
                       bitrate='712k',
                       max_bitrate='712k',
                       buffer_size='1068k',
                       quality='realtime',
                       speed=9,
                       crf=12)


class Vpx360p(BaseVpx):

    def __init__(self, input=None, **opts):
        super(Vpx360p, self).__init__(
            audio=vorbis.LoFiAudio,
            **opts)
        self.codec.set(size='640x360',
                       bitrate='1024k',
                       max_bitrate='1024k',
                       buffer_size='1536k',
                       quality='good',
                       speed=9)


class Vpx480p(BaseVpx):

    def __init__(self, input=None, **opts):
        super(Vpx480p, self).__init__(
            audio=vorbis.MidFiAudio,
            **opts)
        self.codec.set(size='848x480',
                       bitrate='1896k',
                       max_bitrate='1896k',
                       buffer_size='2844k',
                       quality='good',
                       speed=6)


class Vpx576p(BaseVpx):
    def __init__(self, input=None, **opts):
        super(Vpx576p, self).__init__(
            audio=vorbis.MidFiAudio,
            **opts)
        self.codec.set(size='960x576',
                       bitrate='2176k',
                       max_bitrate='2176k',
                       buffer_size='3264k',
                       quality='good',
                       speed=6)


class Vpx720p(BaseVpx):
    def __init__(self, input=None, **opts):
        super(Vpx720p, self).__init__(
            Quantizer(min=0, max=50),
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='1280x270',
                       bitrate='2576k',
                       max_bitrate='2576k',
                       buffer_size='3864k',
                       quality='good')


class Vpx1080p(BaseVpx):
    def __init__(self, input=None, **opts):
        super(Vpx1080p, self).__init__(
            Quantizer(min=0, max=50),
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='1920x1080',
                       bitrate='7552k',
                       max_bitrate='7552k',
                       buffer_size='11328k',
                       crf=8,
                       quality='good')


class Vpx1440p(BaseVpx):
    def __init__(self, input=None, **opts):
        super(Vpx1440p, self).__init__(
            Quantizer(min=0, max=40),
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='2560x1440',
                       bitrate='18M',
                       max_bitrate='18M',
                       buffer_size='30.48M',
                       crf=6,
                       quality='best')


Vpx2k = Vpx1440p


class Vpx2160p(BaseVpx):
    def __init__(self, input=None, **opts):
        super(Vpx2160p, self).__init__(
            Quantizer(min=0, max=40),
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='3840×2160',
                       bitrate='55M',
                       max_bitrate='55M',
                       buffer_size='82.5M',
                       crf=4,
                       quality='best')


Vpx4k = Vpx2160p


class Vp9240p(BaseVp9):

    def __init__(self, input=None, **opts):
        super(Vp9360p, self).__init__(
            audio=opus.LoFiAudio,
            **opts)
        self.codec.set(size='424x240',
                       bitrate='576k',
                       max_bitrate='576k',
                       buffer_size='864k',
                       crf=16,
                       quality='realtime',
                       speed=9)


class Vp9360p(BaseVp9):

    def __init__(self, input=None, **opts):
        super(Vp9360p, self).__init__(
            audio=opus.LoFiAudio,
            **opts)
        self.codec.set(size='640x360',
                       bitrate='896k',
                       max_bitrate='896k',
                       buffer_size='1344k',
                       crf=16,
                       quality='realtime',
                       speed=9)


class Vp9480p(BaseVp9):

    def __init__(self, input=None, **opts):
        super(Vp9480p, self).__init__(
            audio=opus.MidFiAudio,
            **opts)
        self.codec.set(size='848x480',
                       bitrate='1536k',
                       max_bitrate='1536k',
                       buffer_size='2304k',
                       crf=16,
                       quality='good',
                       speed=6)


class Vp9576p(BaseVp9):
    def __init__(self, input=None, **opts):
        super(Vp9576p, self).__init__(
            audio=opus.MidFiAudio,
            **opts)
        self.codec.set(size='960x576',
                       bitrate='2176k',
                       max_bitrate='2176k',
                       buffer_size='3264k',
                       crf=12,
                       quality='good',
                       speed=6)


class Vp9720p(BaseVp9):
    def __init__(self, input=None, **opts):
        super(Vp9720p, self).__init__(
            Quantizer(min=0, max=50),
            audio=opus.HiFiAudio,
            **opts)
        self.codec.set(size='1280x270',
                       bitrate='2576k',
                       max_bitrate='2576k',
                       buffer_size='3864k',
                       quality='good')


class Vp91080p(BaseVp9):
    def __init__(self, input=None, **opts):
        super(Vp91080p, self).__init__(
            Quantizer(min=0, max=50),
            audio=opus.HiFiAudio,
            **opts)
        self.codec.set(size='1920x1080',
                       bitrate='7552k',
                       max_bitrate='7552k',
                       buffer_size='11328k',
                       crf=8,
                       quality='good')


class Vp91440p(BaseVp9):
    def __init__(self, input=None, **opts):
        super(Vp91440p, self).__init__(
            Quantizer(min=0, max=40),
            audio=opus.HiFiAudio,
            **opts)
        self.codec.set(size='2560x1440',
                       bitrate='18M',
                       max_bitrate='18M',
                       buffer_size='30.48M',
                       crf=6,
                       quality='best')


Vp92k = Vp91440p


class Vp92160p(BaseVp9):
    def __init__(self, input=None, **opts):
        super(Vp92160p, self).__init__(
            Quantizer(min=0, max=40),
            audio=opus.HiFiAudio,
            **opts)
        self.codec.set(size='3840×2160',
                       bitrate='55M',
                       max_bitrate='55M',
                       buffer_size='82.5M',
                       crf=4,
                       quality='best')


Vp94k = Vp92160p



from ...codec import H265, MotionEstimation, BFrames, Aac
from ...output import Video, Audio
from ...format import Mp4
from ..audio.aac import (
    LoFiAudio,
    MidFiAudio,
    HiFiAudio,
    LoFiHeAudio,
    MidFiHeAudio,
    HiFiHeAudio
)



class DefaultMe(MotionEstimation):
    def __init__(self, input=None, **opts):
        super(DefaultMe, self).__init__(
            me_method='umh',
            me_range=32,
            subq_est=10,
        )


class LowBFrames(BFrames):

    def __init__(self, input=None, **opts):
        super(LowBFrames, self).__init__(
            strategy=0,
            max=0,
        )


class DefaultBFrames(BFrames):

    def __init__(self, input=None, **opts):
        super(DefaultBFrames, self).__init__(
            strategy=2,
            max=3,
        )


class HighBFrames(BFrames):

    def __init__(self, input=None, **opts):
        super(HighBFrames, self).__init__(
            strategy=2,
            max=5,
        )

class BaseH265(Video):

    def __init__(self, input=None, **opts):
        super(BaseH265, self).__init__(
            DefaultMe(),
            pixel_format='yuv420p',
            codec=H265(rc_lookahead=60,
                       key_interval=120,
                       min_key_interval=1,
                       refs=4,
                       fast_pskip=True,
                       trellis=2,
                       psy_rd=(1.0, 1.0)),
            format=Mp4(),
            **opts)


class AnyDevice(BaseH265):

    def __init__(self, input=None, **opts):
        super(AnyDevice, self).__init__(
            LowBFrames(),
            audio=LoFiAudio,
            **opts)
        self.codec.set(size='424x240',
                       max_bitrate='576k',
                       buffer_size='864k',
                       profile='baseline',
                       level=2.1,
                       partitions=['p4x4'])


class AnyDevice360p(AnyDevice):

    def __init__(self, input=None, **opts):
        super(AnyDevice360p, self).__init__(
            LowBFrames(),
            **opts)
        self.codec.set(size='640x360',
                   level=3.0,
                   max_bitrate='896k',
                   buffer_size='1344k')


class Main360p(BaseH265):

    def __init__(self, input=None, **opts):
        super(Main360p, self).__init__(
            DefaultBFrames(),
            audio=LoFiHeAudio,
            **opts)
        self.codec.set(size='640x360',
                       max_bitrate='896k',
                       buffer_size='1344k',
                       profile='main',
                       level=3.1)


class High360p(Main360p):

    def __init__(self, input=None, **opts):
        super(High360p, self).__init__(**opts)
        self.codec.set(profile='high')


class Main480p(BaseH265):

    def __init__(self, input=None, **opts):
        super(Main480p, self).__init__(
            DefaultBFrames(),
            audio=MidFiHeAudio,
            **opts)
        self.codec.set(size='848x480',
                       max_bitrate='1536k',
                       buffer_size='2304k',
                       profile='main',
                       level=3.1)


class High480p(Main480p):

    def __init__(self, input=None, **opts):
        super(High480p, self).__init__(**opts)
        self.codec.set(profile='high')


class Main576p(BaseH265):
    def __init__(self, input=None, **opts):
        super(Main576p, self).__init__(
            DefaultBFrames(),
            audio=MidFiHeAudio,
            **opts)
        self.codec.set(size='960x576',
                       max_bitrate='2176k',
                       buffer_size='3264k',
                       profile='main',
                       level=3.1)


class High576p(Main576p):
    def __init__(self, input=None, **opts):
        super(High576p, self).__init__(**opts)
        self.codec.set(profile='high')


class High720p(BaseH265):
    def __init__(self, input=None, **opts):
        super(High720p, self).__init__(
            DefaultBFrames(),
            audio=HiFiHeAudio,
            **opts)
        self.codec.set(size='1280x270',
                       max_bitrate='2176k',
                       buffer_size='3264k',
                       profile='high',
                       level=4.0)


class High1080p(BaseH265):
    def __init__(self, input=None, **opts):
        super(High1080p, self).__init__(
            DefaultBFrames(),
            audio=HiFiHeAudio,
            **opts)
        self.codec.set(size='1920x1080',
                       max_bitrate='7552k',
                       buffer_size='11328k',
                       profile='high',
                       level=4.1)


class Super1080p(BaseH265):
    def __init__(self, input=None, **opts):
        super(Super1080p, self).__init__(
            HighBFrames(),
            audio=HiFiHeAudio,
            **opts)
        self.codec.set(size='1920x1080',
                       max_bitrate='12M',
                       buffer_size='18M',
                       profile='high',
                       level=4.1,
                       fast_pskip=False,
                       deblock_alpha=-2,
                       deblock_beta=-1)


class High1440p(BaseH265):
    def __init__(self, input=None, **opts):
        super(High1440p, self).__init__(
            HighBFrames(),
            audio=HiFiHeAudio,
            **opts)
        self.codec.set(size='2560x1440',
                       max_bitrate='18M',
                       buffer_size='30.48M',
                       profile='high',
                       level=4.1,
                       fast_pskip=False,
                       deblock_alpha=-2,
                       deblock_beta=-1)

High2k = High1440p


class High2160p(BaseH265):
    def __init__(self, input=None, **opts):
        super(High2160p, self).__init__(
            HighBFrames(),
            audio=HiFiHeAudio,
            **opts)
        self.codec.set(size='3840×2160',
                       max_bitrate='55M',
                       buffer_size='82.5M',
                       profile='high',
                       level=4.1,
                       fast_pskip=False,
                       deblock_alpha=-2,
                       deblock_beta=-1)

High4k = High2160p



# -*- coding: utf-8 -*-
import io
import pathlib
from awstools import S3Manager
from vital.cache import cached_property
from ..cli import *
from ._protocol import Protocol
from .pipe import Pipe
from .file import FileIn


__all__ = 'S3In', 'S3Out', 'S3BytesIn', 'S3FileIn', 'S3OptimalIn'


class _S3(Pipe):
    """
    `S3 input/output protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
    @bucket: (#str) name of the S3 bucket to pull the file from or put the
        file to
    @filename: (#str) name of the file to pull or output to and from S3
    @put_opt: (#dict) keyword arguments to pass to boto.client.put_object
    """
    NUMBER = 0

    def __init__(self, *opt, **opts):
        """ @endpoint: (#str) e.g. s3://your.bucket/path/to/vid.mp4 """
        super().__init__(*opt, **opts)
        self.manager = S3Manager()

    @staticmethod
    def parse_endpoint(endpoint):
        endpoint = endpoint.split('s3://', 1)[1]
        return endpoint.split('/', 1)


class S3In(_S3):
    NUMBER = 0

    def __init__(self, endpoint, *opt, buffer_size=1024 * 1024 * 24, **opts):
        """ @endpoint: (#str) e.g. s3://your.bucket/path/to/vid.mp4 """
        self.bucket, self.filename = self.parse_endpoint(endpoint)
        self.buffer = buffer_size
        super().__init__(*opt, **opts)

    @cached_property
    def body(self):
        return self.manager.get_file(self.bucket, self.filename).get()['Body']

    def iter_body(self):
        while True:
            chunk = self.body.read(self.buffer)
            if not len(chunk):
                break
            yield chunk


class S3FileIn(FileIn):
    NUMBER = 0

    def __init__(
        self,
        s3_endpoint,
        local_path,
        *opt,
        **opts
    ):
        """ @endpoint: (#str) e.g. s3://your.bucket/path/to/vid.mp4 """
        self.bucket, filename = _S3.parse_endpoint(s3_endpoint)
        self.path = pathlib.Path(local_path)
        self.path.parent.mkdir(0o744, parents=True, exist_ok=True)

        super().__init__(str(self.path), *opt, **opts)
        self.manager = S3Manager()
        self.manager.client.download_file(self.bucket, filename, str(self.path))

    def close(self):
        self.path.unlink()


class S3BytesIn(S3In):
    NUMBER = 0

    @cached_property
    def body(self):
        body = io.BytesIO()
        self.manager.client.download_fileobj(self.bucket, self.filename, body)
        body.seek(0)
        return body

    def iter_body(self):
        self.body.seek(0)

        while True:
            chunk = self.body.read(self.buffer)
            if not len(chunk):
                break
            yield chunk


def S3OptimalIn(
    s3_endpoint,
    local_path,
    *args,
    bytes_thresh=1024 * 1024 * 24,
    file_thresh=1024 * 1024 * 240,
    **kwargs
):
    s3 = _S3()
    bucket, key = s3.parse_endpoint(s3_endpoint)
    response = s3.manager.client.head_object(Bucket=bucket, Key=key)
    size = response['ContentLength']

    if size >= bytes_thresh:
        if size >= file_thresh:
            print(f'*** Writing S3 to file first [{local_path}] ***')
            return S3FileIn(s3_endpoint, local_path, *args, **kwargs)
        else:
            print('*** Writing S3 to BytesIO object first ***')
            return S3BytesIn(s3_endpoint, *args, **kwargs)
    else:
        print('*** Streaming S3 through pipe ***')
        return S3In(s3_endpoint, *args, **kwargs)


class S3Out(_S3):
    NUMBER = 1

    def __init__(self, endpoint, put_opt=None, *opt, **opts):
        """ @endpoint: (#str) e.g. s3://your.bucket/path/to/vid.mp4 """
        self.bucket, self.filename = self.parse_endpoint(endpoint)
        self.put_opt = put_opt
        super().__init__(*opt, **opts)

    def put(self, body):
        return self.manager.upload_stream(
            self.bucket,
            self.filename,
            body,
            **self.put_opt
        )


class S3Pipe(_S3):
    pass



# -*- coding: utf-8 -*-
from ...cli import *
from .._container import *


__all__ = 'Hls',


# https://ffmpeg.org/ffmpeg-formats.html#hls-2


class Hls(Container):
    """
    Apple HTTP Live Streaming muxer that segments MPEG-TS according to the
    HTTP Live Streaming (HLS) specification.

    It creates a playlist file, and one or more segment files. The output
    filename specifies the playlist filename.

    By default, the muxer creates a file for each segment produced. These files
    have the same name as the playlist, followed by a sequential number and a
    .ts extension.


    ``HLS muxer options``
    ~~~~~~~~~~~~~~~~~~~~~
    @time (hls_time): (#float) Set the segment length in seconds. Default value
        is 2.
    @list_size (hls_list_size): (#int) Update the list file so that it contains
        at most size segments. If 0 the list file will contain all the segments.
        Default value is 0.
    @wrap (hls_wrap): (#int) Wrap around segment index once it reaches
        limit.
    @ts_options (hls_ts_options): (#dict) Set output format options.
    @start_number: (#int) Set the sequence number of the
        first segment. Defaults to 0.
    @allow_cache (hls_allow_cache): (#bool) Explicitly set whether the client
        MAY (1) or MUST NOT (0) cache media segments.
    @base_url (hls_base_url): (#str) Append baseurl to every entry in the
        playlist. Useful to generate playlists with absolute paths.
    @segment_filename (hls_segment_filename): (#str) Set the segment filename.
        Unless hls_flags single_file is set filename is used as a string format
        with the segment number:
    @use_localtime_mkdir: (#bool) Used together with -use_localtime, it will
        create up to one subdirectory which is expanded in filename.
    @key_info_file (hls_key_info_file): (#str) Use the information in
        key_info_file for segment encryption. The first line of key_info_file
        specifies the key URI written to the playlist. The key URL is used to
        access the encryption key during playback. The second line specifies
        the path to the key file used to obtain the key during the encryption
        process. The key file is read as a single packed array of 16 octets in
        binary format. The optional third line specifies the initialization
        vector (IV) as a hexadecimal string to be used instead of the segment
        sequence number (default) for encryption. Changes to key_info_file will
        result in segment encryption with the new key/IV and an entry in the
        playlist for the new key URI/IV.
    @hls_flags: (flags)
    @playlist_type (hls_playlist_type): (#str) 'vod' or 'event'
    @init_filename: (#str) set filename to the fragment files header file,
        default filename is init.mp4.
    """
    OPT = [
        Opt('f', 'format', value='hls'),
        FloatOpt('hls_time', 'time'),
        IntOpt('hls_list_size', 'list_size'),
        KeyValOpt('hls_ts_options', 'ts_options'),
        IntOpt('start_number'),
        OneOfOpt(
            'hls_start_number_source',
            'start_number_source',
            choices=('generic', 'epoch', 'datetime')
        ),
        EnableOpt('hls_allow_cache', 'allow_cache'),
        Opt('hls_base_url', 'base_url'),
        Opt('hls_segment_filename', 'segment_filename'),
        Flag('use_localtime'),
        Flag('use_localtime_mkdir'),
        Opt('hls_key_info_file', 'key_info_file'),
        FlagsOpt('hls_flags'),
        OneOfOpt(
            'hls_playlist_type',
            'playlist_type',
            choices=('vod', 'event')
        ),
        EnableOpt('hls_enc', 'encrypt'),
        Opt('hls_enc_key', 'encrypt_key'),
        Opt('hls_enc_key_url', 'encrypt_key_url'),
        Opt('hls_enc_iv', 'encrypt_iv'),
        OneOfOpt(
            'hls_segment_type',
            'segment_type',
            choices=('fmp4', 'mpegts')
        ),
        Opt('hls_fmp4_init_filename', 'init_filename'),
        OneOfOpt('method', ('PUT', 'GET', 'POST', 'DELETE', 'HEAD')),
        Opt('user_agent')
    ]



# -*- coding: utf-8 -*-
from ..cli import *
from ..probe import Probe


__all__ = (
    'Codec',
    'VideoCodec',
    'AudioCodec',
    'Quantizer',
    'BFrames',
    'IFrames',
    'CompareFunctions',
    'Macroblocks',
    'MotionEstimation'
)


DEFAULT_CODEC_OPT = [
    Opt('c', 'codec'),
    BitOpt('b', 'bitrate'),
    IntOpt('threads', default=0),
    BitOpt('maxrate', 'max_bitrate'),
    BitOpt('minrate', 'min_bitrate'),
    BitOpt('bufsize', 'buffer_size'),
    IntOpt('g', 'key_interval'),
    IntOpt('keyint_min', 'min_key_interval'),
    BitOpt('bt', 'bitrate_tolerance'),
    FlagsOpt('flags'),
    FlagsOpt('flags2'),
    IntOpt('frame_number'),
    OneOfOpt('strict', choices=('very', 'strict', 'normal',
                                'experimental')),
    OneOfOpt('err_detect',
             choices=('crccheck', 'bitstream', 'buffer', 'explode',
                      'ignore_err', 'careful', 'compliant',
                      'aggressive')),
    OneOfOpt('ec', 'error_concealment',
             choices=('guess_mvs', 'deblock', 'favor_inter')),
    IntOpt('bits_per_coded_sample'),
    IntOpt('rc_init_occupancy'),
    IntOpt('lowres'),
    IntOpt('skip_threshold'),
    IntOpt('skip_factor'),
    IntOpt('skip_exp'),
    IntOpt('slices'),
    Opt('sub_charenc')
]


class Codec(OptSet):
    '''
    -threads #int Use 0 for auto
    -maxrate #int (bit/s) Set the max bitrate tolerance. Requires -bufsize
                  option to be set.
    -minrate #int Set minimum bitrate tolerance (bit/s). Most useful in setting
                  up a CBR encode, otherwise it is of little use.
    -bufsize #int Set ratecontrol buffer size (in bits)
    -bt - bitrate tolerance in bit/s
    -flags flags (mv4|qpel|loop|qscale|gmc|mv0|input_preserved|pass1|pass2|gray|
                  emu_edge|psnr|truncated|naq|ildct|low_delay|global_header|
                  bitexact|aic|ilme|cgop)
    -frame_number #int Set the frame number
    -strict #int Specify how strictly to follow standards
                 (very|strict|normal|experimental)
    -err_detect Set error detection flags. (crccheck|bitstream|buffer|explode|
                ignore_err|careful|compliant|aggressive)
    -ec Set error concealment strategy (guess_mvs|deblock|favor_inter)
    -bits_per_coded_sample #int
    -rc_init_occupancy #int Set number of bits which should be loaded into the
                       rc buffer before decoding starts.
    -flags2 (fast|noout|ignorecrop|local_header|chunks|showall|expor_mvs)
    -profile #int
    -level #int
    -lowres #int Decode at 1 = 1/2, 2=1/4, 3=1/8 resolutions
    -skip_threshold #int Set frame skip threshold
    -skip_factor #int Set frame skip factor
    -skip_exp #int Skip frame exponent
    -slices #int Number of slices, used in parallelized encoding
    -sub_charenc Sets the input subtitles character encoding.
    -keyint_min #int Set minimum interval between IDR-frames
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_CODEC_OPT + self.OPT + list(opt)), **opts)

    def __call__(self, *args, **kwargs):
        return self.__class__(*args, **kwargs)


DEFAULT_VIDEO_OPT = [
    Opt('c:v', 'codec'),
    BitOpt('b:v', 'bitrate'),
    Opt('profile:v', 'profile'),
    Opt('level:v', 'level'),
    FloatOpt('r', 'frame_rate'),
    SizeOpt('s', 'size'),
    ByteOpt('ps', 'rtp_payload_size'),
    IntOpt('lelim', 'luminance_elim'),
    IntOpt('celim', 'chrominance_elim'),
    OneOfOpt(
        'dct',
        choices=(
            'auto',
            'fastint',
            'mmx',
            'altivec',
            'faan'
        )
    ),
    FloatOpt('rc_init_cplx', 'init_complexity'),
    FloatOpt('lumi_max', 'max_luminance'),
    FloatOpt('tcplx_mask', 'temporal_complexity'),
    FloatOpt('scplx_mask', 'spatial_complexity'),
    FloatOpt('p_mask', 'inter_masking'),
    FloatOpt('dark_mask', 'dark_masking'),
    IntOpt('idct', 'idct'),
    OneOfOpt(
        'pred',
        'prediction_method',
        choices=('left', 'plane', 'median')
    ),
    FloatOpt('aspect'),
    IntOpt('ibias', 'intra_quant_bize'),
    OneOfOpt(
        'coder',
        choices=('vlc', 'ac', 'raw', 'rle', 'deflate')
    ),
    IntOpt('context'),
    IntOpt('sc_threshold', 'scene_change_thresh'),
    IntOpt('lmin', 'min_lagrange'),
    IntOpt('lmax', 'max_lagrange'),
    IntOpt('nr', 'noise_reduction'),
    IntOpt('refs'),
    IntOpt('chromaoffset'),
    IntOpt('trellis'),
    IntOpt('sc_factor', 'scene_change_factor'),
    IntOpt('colorspace'),
    IntOpt('dc', 'dc_precision'),
    IntOpt('nssew', 'nsse_weight'),
    OneOfOpt(
        'field_order',
        choices=('progressive', 'tt', 'bb', 'tb', 'bt')
    ),
    OneOfOpt('skip_alpha', cast=int, choices=(0, 1))
]


class VideoCodec(Codec):
    '''
    -g #int Set the GOP size. (default is 12)
        (video)
    -ps #int Set RTP payload size in bytes
    -lelim #int Set single coefficient elimination threshold for luminance
    -celim #int Set single coefficient elimination threshold for chrominance
    -dct #int Set DCT algorithm (auto|fastint|int|mmx|altivec|faan)
    -rc_init_cplx #float Set initial complexity for 1-pass encoding.
    -lumi_max #float Compress bright areas stronger than medium was
    -tcplx_max #float Set temporal complexity masking
    -scplx_mask #float Set spatial complexity masking
    -p_mask #float Set inter masking
    -dark_mask #float Compress dark areas stronger than medium ones
    -idct #int Select IDCT implementation
    -pred #int Set prediction method. (left|plane|median)
    -aspect #float Set sample aspect ratio
    -last_pred #int Set amount of motion predictors from the previous frame.
    -coder #int (vlc|ac|raw|rle|deflate)
    -context #int Set context model
    -sc_threshold #int Set scene change threshold
    -lmin #int Set min lagrange factor
    -lmax #int Set max lagrange factor
    -nr #int Set noise reduction
    -refs #int Set reference frames to consider for motion compensation
    -chromaoffset #int Set chroma qp offset from luma
    -trellis #int Set rate-distortion optimal quantization.
    -sc_factor #int Set value multiplied by qscale for each from and added
                to scene_change_score
    -colorspace #int
    -dc #int Set intra_dc_precision
    -nssew #int Set nsse weight
    -field_order Set/override the field order of the video.
                (progressive|tt|bb|tb|bt)
    -skip_alpha #int Set to 1 to disable processing alpha (transparency)
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_VIDEO_OPT + self.OPT + list(opt)), **opts)


DEFAULT_AUDIO_OPT = [
    Opt('c:a', 'codec'),
    Opt('profile:a', 'profile'),
    Opt('level:a', 'level'),
    Opt('b:a', 'bitrate'),
    IntOpt('ar', 'sample_rate'),
    IntOpt('ac', 'channels'),
    IntOpt('aq', 'quality'),
    IntOpt('cutoff'),
    IntOpt('frame_size'),
    OneOfOpt(
        'audio_service_type',
        'service_type',
        choices=('ma', 'ef', 'vi', 'hi', 'di', 'co', 'em', 'vo', 'ka')
    )
]


class AudioCodec(Codec):
    '''
    -ar #int Set audio sampling rate in Hz
    -ac #int Set the number of audio channels
    -aq #int Set audio quality (codec-specific)
    -cutoff #int Set the cutoff bandwidth
    -frame_size #int Set the audio frame size. Each submitted frame
                except the last must contain exactly frame_size samples per
                channel
    -audio_service_type #int Set audio service type.
                        (ma|ef|vi|hi|di|co|em|vo|ka)
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_AUDIO_OPT + self.OPT + list(opt)), **opts)


DEFAULT_QUANTIZER_OPT = [
    BoundedOpt('qcomp', 'curve_compression', lower=0.0, upper=1.0, cast=float),
    FloatOpt('qblur', 'blur'),
    IntOpt('qmin', 'min'),
    IntOpt('qmax', 'max'),
    IntOpt('qdiff', 'max_diff'),
    IntOpt('ibias'),
    IntOpt('pbias')
]


class Quantizer(OptSet):
    '''
    -qcomp #float Quantizer-Curve Compression
    -qblur #float Set the video quantizer scale blur (VBR)
        (video)
    -qmin #int Set the minimum video quantizer scale (VBR) Must be included
          between -1 and 69, default value is 2.
        (video)
    -qmax #int Set max video quantizer scale (VBR). Must be included between
          -1 and 1024, default value is 31.
        (video)
    -qdiff #int Set the max difference between the quantizer scale (VBR)
        (video)
    -ibias #int Set intra quant bias
    -pbias #int Set inter quant bias
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_QUANTIZER_OPT + self.OPT + list(opt)), **opts)


DEFAULT_BFRAMES_OPT = [
    BoundedOpt('bf', 'max', lower=-1, upper=16, default=0),
    FloatOpt('b_qfactor', 'q_factor'),
    FloatOpt('b_qoffset', 'qp_offset'),
    OneOfOpt('b_strategy', 'strategy', cast=int, choices=(0, 1, 2))
]


class BFrames(OptSet):
    '''
    -bf #int Set the max number of B frames between non-B-frames. Must be an
        integer between -1 and 16. 0 means that B-frames are disabled. If a
        value of -1 is used, it will choose an automatic value depending on
        the encoder. Default is 0.
        (video)
    -b_qfactor #float Set qpfactor between P and B frames.
        (video)
    -b_qoffset #float Set QP offset between P and B-frames
        (video)
    -b_strategy #int Set strategy to choose between I/P/B-frames
        (video)
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_BFRAMES_OPT + self.OPT + list(opt)), **opts)


DEFAULT_IFRAMES_OPT = [
    FloatOpt('i_qfactor', 'q_factor'),
    FloatOpt('i_qoffset', 'qp_offset')
]


class IFrames(OptSet):
    '''
    -i_qfactor #float Set QP factor between P and I frames
        (video)
    -i_qoffset #float Set QP offset between P and I frames
        (video)
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_IFRAMES_OPT + self.OPT + list(opt)), **opts)


CF_CHOICES = (
    'sad', 'sse', 'satd', 'dct', 'psnr', 'bit', 'rd', 'zero',
    'vsad', 'vss3', 'nss3', 'wv53', 'w97', 'dctmax', 'chroma'
)

DEFAULT_CF_OPT = [
    OneOfOpt('cmp', 'pel_me', choices=CF_CHOICES),
    OneOfOpt('subcmp', 'sub_pel_me', choices=CF_CHOICES),
    OneOfOpt('mbcmp', 'macroblock_cmp', choices=CF_CHOICES),
    OneOfOpt('ildctcmp', 'interlaced_dct', choices=CF_CHOICES),
    OneOfOpt('precmp', 'pre_motion_est', choices=CF_CHOICES),
    OneOfOpt('skipcmp', 'skip', choices=CF_CHOICES)
]


class CompareFunctions(OptSet):
    '''
    -cmp #int Set full pel me compare function (sad(default)|sse|satd|dct|psnr|
              bit|rd|zero|vsad|vss3|nss3|wv53|w97|dctmax|chroma)
    -subcmp #int Set sub pel me compare function (sad(default)|sse|satd|dct|
                 psnr|bit|rd|zero|vsad|vss3|nss3|wv53|w97|dctmax|chroma)
    -mbcmp #int Set macroblock compare function (sad(default)|sse|satd|dct|
                psnr|bit|rd|zero|vsad|vss3|nss3|wv53|w97|dctmax|chroma)
    -ildctcmp #int Set interlaced dct compare function (sad(default)|sse|satd|
                   dct|psnr|bit|rd|zero|vsad|vss3|nss3|wv53|w97|dctmax|chroma)
    -precmp #int Set the pre motion estimation compare function
                 (sad(default)|sse|satd|dct|psnr|bit|rd|zero|vsad|vss3|nss3|
                  wv53|w97|dctmax|chroma)
    -skipcmp #int (sad(default)|sse|satd|dct|psnr|bit|rd|zero|vsad|vss3|nss3|
             wv53|w97|dctmax|chroma)
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_CF_OPT + self.OPT + list(opt)), **opts)


DEFAULT_MB_OPT = [
    OneOfOpt('mbd', 'decision_algo', choices=('simple', 'bits', 'rd')),
    IntOpt('mblmin', 'min'),
    IntOpt('mblmax', 'max'),
    IntOpt('mb_threshold', 'threshold'),
    IntOpt('skip_top'),
    IntOpt('skip_bottom'),
    FloatOpt('border_mask')
]


class Macroblocks(OptSet):
    '''
    -mbd #int Set macroblock decision algorithm (high quality mode)
              (simple|bits|rd)
    -mblmin #int Set min macroblock lagrange factor (VBR)
    -mblmax #int Set max macroblock lagrange factor (VBR)
    -mb_threshold #int Set macroblock threshold.
    -skip_top #int Set number of macroblock rows at the top which are skipped.
    -skip_bottom #int Set number of macroblock rows at the bottom which are
                 skipped.
    -border_mask #float Increase the quantizer for macroblocks close to borders.
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_MB_OPT + self.OPT + list(opt)), **opts)


DEFAULT_ME_OPT = [
    OneOfOpt(
        'me_method',
        choices=(
            'zero',
            'full',
            'epzs',
            'esa',
            'tesa',
            'dia',
            'log',
            'phods',
            'x1',
            'hex',
            'umh',
            'iter'
        )
    ),
    IntOpt('me_range'),
    IntOpt('dia_size'),
    IntOpt('pre_dia_size'),
    IntOpt('preme', 'pre'),
    IntOpt('subq', 'subq_est'),
    BoundedOpt('mepc', 'bitrate_penalty_comp', lower=1, upper=256),
    IntOpt('me_threshold', 'threshold')
]


class MotionEstimation(OptSet):
    '''
    -me_method Set motion estimation method. One of the most important
               settings for x264, both speed and quality-wise.
               (zero|full|epzs|esa|tesa|dia|log|phods|x1|hex|umh|iter)
    @me_range: (#int) MErange controls the max range of the motion search.
        For HEX and DIA, this is clamped to between 4 and 16, with a default
        of 16. For UMH and ESA, it can be increased beyond the default 16 to
        allow for a wider-range motion search, which is useful on HD footage
        and for high-motion footage. Note that for UMH and ESA, increasing
        MErange will significantly slow down encoding.
    -dia_size #int Set diamond type and size for motion estimation
    -pre_dia_size #int Set diamond type and size for motion estimation pre-pass
    -preme #int Set the pre motion estimation
    -subq #int Set sub pel motion estimation quality
    -mepc #int Set motion estimation bitrate penalty compensation (1.0 = 256)
    -me_threshold #int Set motion estimation threshold.
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_ME_OPT + self.OPT + list(opt)), **opts)



from ...codec import *


#: Audio presets
LoFiAudio = Vorbis(bitrate='64k', sample_rate='44100', channels=2)
MidFiAudio = Vorbis(bitrate='128k', sample_rate='44100', channels=2)
HiFiAudio = Vorbis(bitrate='320k', sample_rate='44100')



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = ('Aiff',)


class Aiff(Container):
    """
    ``Audio Interchange File Format container format``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @write_id3v2: (#bool) Enable ID3v2 tags writing when set to 1.
        Default is 0 (disabled).
    @id3v2_version: (#int) Select ID3v2 version to write. Currently only
        version 3 and 4 (aka. ID3v2.3 and ID3v2.4) are supported. The default
        is version 4.

    """
    OPT = [
        Opt('f', 'format', value='aiff'),
        EnableOpt('write_id3v2'),
        OneOfOpt('id3v2_version', cast=int, choices=(3, 4))
    ]



# -*- coding: utf-8 -*-
'''
                #####
                 ##
                 #
    `Transcend` #
    ~~~~~~~~~~~~~~~~~
    High performance video and audio transcoding for aliens.


    usage: ts s3://queue.bucket:v/big-buck-bunny.mp4 \
              s3://public.bucket:v/big-buck-bunny.mp4 \
              --budget 0.35 \
              --max-filesize 150M


    python usage:

    import transcend as ts
    audio_track = ts.Audio(codec=ts.Aac(), channels=2)
    output = ts.Video(codec=ts.H264(profile='high',
                                    level=4.1,
                                    size=(1280, 720)),
                      container=ts.format.Mp4(),
                      audio=audio_track)

    Things to remember
    ==================

    * bitrate = file size / duration
        * allow setting desired or max file size
        * 24 MBPS * (2 hours * 60 minutes / hour * 60 seconds / minute) = 172,800 MB
        ------------------------
        frames per sec. - 30
        res.width - 1920
        res.height - 1080
        Gop size - 12
        frame/sec by Gopsize  (30/12=2,5)
        pixels in 1 frame  (1920*1080=2.073.600)
        pixels in frame/sec  (2.073.600*30=62.208.000)
        bitrate max needed  (62.208.000/2,5=24.883.200) -24000 - 24Mb/s
        bitrate with 70% loses  (24.883.200/0,7=17.418.240) - 17000 - 17Mb/s
        For 720i with loses 70% and will give a size of file equalling 10Gb
        for losses of 40% give 4Gb and screens with a res of 720-480 and an
        aspect ratio of 3/4, this will give size 3,7Gb for 70% and
        1,4Gb for 40%
    * audio bitrate
        * fileSize = (bitsPerSample * samplesPerSecond * channels * duration) / 8;
            * 30 seconds of stereo will take up (8 * 44100 * 2 * 30) / 8 = 2,646,000
    * optimize for -> encoding speed OR best compression


    Cheat sheet
    ===========
    http://www.labnol.org/internet/useful-ffmpeg-commands/28490/
    http://www.catswhocode.com/blog/19-ffmpeg-commands-for-all-needs
    https://blogs.dropbox.com/tech/2014/02/video-processing-at-dropbox/
    http://www.lighterra.com/papers/videoencodingh264/
'''

# -*- coding: utf-8 -*-
from __future__ import print_function, division
from .format import *
from .preset import *
from .protocol import *
from .output import *
from .ffmpeg import *
from .probe import *
from . import codec
from . import preset
from . import protocol
from . import faststart
from . import utils


if __name__ == '__main__':
    from vital.debug import Timer
    import transcend as ts
    t = Timer()
    t.start()
    f = ts.Ffmpeg(loglevel='error', overwrite=True)
    # f.input = ts.protocol.S3In('s3://xfaps.queue/v/3d/3d62ba17-ef93-4c75-81af-bf8604b445ea.mp4')
    f.input = ts.protocol.S3In('s3://xfaps.queue/Sintel.2010.1080p.mkv')
    # f.input = ts.protocol.S3In('s3://xfaps.queue/v/42/42e2bc9c-eb23-44b7-b870-1aba196b1172.mp4')
    # f.save(ts.preset.h264.AnyDevice360p(output=ts.protocol.FileOut('test_files/test.mp4')))
    # f.save(ts.preset.vpx.AnyDevice360p(output=ts.protocol.FileOut('test_files/test.vpx.webm')))
    f.save(ts.preset.dash.H264xAnyDevice(output=ts.protocol.FileOut('test_files/test.240p.m4v')),
           ts.preset.dash.H264xAnyDevice360p(output=ts.protocol.FileOut('test_files/test.360p.m4v')),
           ts.preset.dash.H264x576p(output=ts.protocol.FileOut('test_files/test.576p.m4v')),
           ts.preset.dash.H264x720p(output=ts.protocol.FileOut('test_files/test.720p.m4v')),
           ts.Audio(disable_video=True,
                    disable_subtitles=True,
                    codec=ts.preset.dash.LoFiDashAudio.__class__(flags='cgop', key_interval=120, min_key_interval=120),
                    format=ts.format.Mp4(movflags='frag_keyframe+faststart'),
                    output=ts.protocol.FileOut('test_files/test.audio.hev2.m4a')),
            ts.Audio(disable_video=True,
                     disable_subtitles=True,
                     codec=ts.preset.dash.HiFiDashAudio.__class__(flags='cgop', key_interval=120, min_key_interval=120),
                     format=ts.format.Mp4(movflags='frag_keyframe+faststart'),
                     output=ts.protocol.FileOut('test_files/test.audio.hev2.hi.m4a')))
    print(t.stop())



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = ('SftpIn', 'SftpOut')


SFTP_OPT = [IntOpt('timeout'), EnableOpt('truncate'), EnableOpt('private_key ')]


class _Sftp(Protocol):
    """
    `Secure file transfer protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @timeout: (#int) Set timeout in microseconds of socket I/O operations used
        by the underlying low level operation. By default it is set to -1,
        which means that the timeout is not specified.
    @truncate: (#bool) Truncate existing files on write, if set to 1. A value
        of 0 prevents truncating. Default value is 1.
    @private_key: (#str) Specify the path of the file containing private key to
        use during authorization. By default libssh searches for keys in the
        ~/.ssh/ directory.
    """
    OPT = SFTP_OPT


class SftpIn(_Sftp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class SftpOut(_Sftp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Mp4',


class Mp4(Container):
    """
    The mov/mp4/ismv muxer supports fragmentation. Normally, a MOV/MP4 file has
    all the metadata about all packets stored in one location (written at the
    end of the file, it can be moved to the start for better playback by adding
    faststart to the movflags, or using the qt-faststart tool). A fragmented
    file consists of a number of fragments, where packets and metadata about
    these packets are stored together. Writing a fragmented file has the a
    dvantage that the file is decodable even if the writing is interrupted
    (while a normal MOV/MP4 is undecodable if it is not properly finished),
    and it requires less memory when writing very long files (since writing
    normal MOV/MP4 files stores info about every single packet in memory until
    the file is closed). The downside is that it is less compatible with other
    applications.


    `Mp4 container format`
    ~~~~~~~~~~~~~~~~~~~~~~
    @moov_size: (#str) Reserves space for the moov atom at the beginning of the
        file instead of placing the moov atom at the end. If the space reserved
        is insufficient, muxing will fail.
    @frag_duration: (#int) Create fragments that are duration microseconds long.
    @frag_size: (#str) Create fragments that contain up to size bytes of
        payload data.
    @min_frag_duration: (#int) Don’t create fragments that are shorter than
        duration microseconds long.
    """
    OPT = [
        Opt('f', 'format', value='mp4'),
        ByteOpt('moov_size'),
        # FlagsOpt('movflags', value='+faststart'),
        FlagsOpt('movflags'),
        IntOpt('frag_duration'),
        ByteOpt('frag_size'),
        IntOpt('min_frag_duration')
    ]



# -*- coding: utf-8 -*-
from ...cli import *
from .._protocol import Protocol


__all__ = 'NullOut',


class NullOut(Protocol):
    """
    `/dev/null output protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    """
    def __init__(self, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value='/dev/null'),
            **opts
        )



from ...codec import *


#: Audio presets
LoFiAudio = Aac(bitrate='64k', sample_rate='44100', channels=2)
MidFiAudio = Aac(bitrate='128k', sample_rate='44100', channels=2)
HiFiAudio = Aac(bitrate='320k', sample_rate='44100')


#: HE-AAC audio presets
LoFiHeAudio = Aac(bitrate='24k', sample_rate='44100', channels=2,
                  profile='aac_he_v2')
# Not all devices can use he-aac audio but we should expect hi fi devices to be
# be able to at this point
MidFiHeAudio = Aac(bitrate='48k', sample_rate='44100', channels=2,
                   profile='aac_he_v2')
HiFiHeAudio = Aac(bitrate='160k', sample_rate='44100', profile='aac_he')



# -*- coding: utf-8 -*-
from ...cli import *
from .._protocol import Protocol


__all__ = 'Md5Out',


class Md5Out(Protocol):
    """
    `MD5 output protocol.`
    ~~~~~~~~~~~~~~~~~~~~~~~~
    Computes the MD5 hash of the data to be written, and on close writes this
    to the designated output or stdout if none is specified. It can be used to
    test muxers without writing an actual file.
    """
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value='md5:' + endpoint),
            **opts
        )
